/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 * 
 * Modified by Matthieu Chaffotte - BonitaSoft S.A.
 **/
package org.ow2.bonita.connector.core;

import java.io.File;
import java.lang.reflect.Modifier;
import java.net.URI;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.List;
import java.util.Locale;
import java.util.ResourceBundle;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

import org.ow2.bonita.connector.impl.mapper.PerformerAssignFilter;

/**
 * 
 * @author chaffotm
 *
 */
public class ConnectorAPI {

  private List<ConnectorDescription> connectors;
  private Locale currentLocale;
  

  @Deprecated
  public ConnectorAPI(URI uri) throws Exception {
    this(uri, Locale.ENGLISH);
  }

  @Deprecated
  public ConnectorAPI(URI uri, Locale locale) throws Exception {
    this.currentLocale = locale;
    this.connectors = new ArrayList<ConnectorDescription>();
    loadConnectors(uri);
  }
  
  @Deprecated
  private void loadConnectors(URI uri) throws Exception {
    File file = new File(uri);
    loadConnectors(new JarFile(file));
  }


  @Deprecated
  public ConnectorAPI(JarFile jarFile) throws Exception {
    this(jarFile, Locale.ENGLISH);
  }

  @Deprecated
  public ConnectorAPI(JarFile jarFile, Locale locale) throws Exception {
    this.currentLocale = locale;
    this.connectors = new ArrayList<ConnectorDescription>();
    loadConnectors(jarFile);
  }
  
  public ConnectorAPI(ClassLoader classLoader, List<String> classNames) throws Exception {
    this(classLoader, classNames, Locale.getDefault());
  }

  public ConnectorAPI(ClassLoader classLoader, List<String> classNames, Locale locale) throws Exception {
    this.currentLocale = locale;
    this.connectors = new ArrayList<ConnectorDescription>();
    loadConnectors(classLoader, classNames);
  }
  
  public void loadConnectors(ClassLoader classLoader, List<String> classNames) throws Exception {
    List<Class<?>> classes = new ArrayList<Class<?>>();
    for (String className : classNames) {
      classes.add(classLoader.loadClass(className));
    }
    loadConnectors(classes);
  }

  @Deprecated
  @SuppressWarnings("unchecked")
  private void loadConnectors(JarFile jarFile) throws Exception {
    Enumeration<JarEntry> files = jarFile.entries();
    List<ConnectorDescription> tempList = new ArrayList<ConnectorDescription>();
    while (files.hasMoreElements()) {
      JarEntry file = files.nextElement();
      if (!file.isDirectory()) {
        String fileName = file.getName();
        if (fileName.endsWith(".class")) {
          fileName = fileName.replace('/', '.');
          int length = fileName.length();
          String className = fileName.substring(0, length - 6);
          if (!className.contains("$") && className.startsWith("org.ow2.bonita.connector.impl")) {
            Class<?> clazz = Class.forName(className);
            if (isAnInstanceOfConnector(clazz)) {
              ConnectorDescription temp =
                new ConnectorDescription((Class<? extends Connector>)clazz, currentLocale);
              String id = temp.getId();
              if (idExists(id)) {
                throw new IllegalArgumentException("Cannot load " + clazz.getName()
                    + " because a similar id exists");
              }
              tempList.add(temp);
            }
          }
        }
      }
    }
    connectors.addAll(tempList);
  }
  
  @SuppressWarnings("unchecked")
  private void loadConnectors(List<Class<?>> classes) {
    List<ConnectorDescription> tempList = new ArrayList<ConnectorDescription>();
    for (Class<?> clazz : classes) {
      if (isAnInstanceOfConnector(clazz) && !clazz.equals(PerformerAssignFilter.class)) {
        ConnectorDescription temp =
          new ConnectorDescription((Class<? extends Connector>)clazz, currentLocale);
        String id = temp.getId();
        if (idExists(id)) {
          throw new IllegalArgumentException("Cannot load " + clazz.getName()
              + " because a similar id exists");
        }
        tempList.add(temp);
      }
    }
    connectors.addAll(tempList);
  }

  private boolean idExists(String id) {
    boolean exists = false;
    for (ConnectorDescription connector : connectors) {
      if (id.equals(connector.getId())) {
        exists = true;
        break;
      }
    }
    return exists;
  }

  public List<ConnectorDescription> getAllConnectors() {
    return connectors;
  }

  public List<ConnectorDescription> getJavaConnectors() {
  	List<ConnectorDescription> javaconnectors = new ArrayList<ConnectorDescription>();
  	for (ConnectorDescription connector : connectors) {
  		Class<? extends Connector> clazz = connector.getConnectorClass();
  		if (isConnector(clazz) && !isMapper(clazz)) {
  			javaconnectors.add(connector);
  		}
    }
  	return javaconnectors;
  }

  public List<ConnectorDescription> getRoleResolvers() {
  	List<ConnectorDescription> roleResolvers = new ArrayList<ConnectorDescription>();
  	for (ConnectorDescription connector : connectors) {
  		Class<? extends Connector> clazz = connector.getConnectorClass();
  		if (isRoleResolver(clazz)) {
  			roleResolvers.add(connector);
  		}
    }
  	return roleResolvers;
  }

  public List<ConnectorDescription> getFilters() {
  	List<ConnectorDescription> filters = new ArrayList<ConnectorDescription>();
  	for (ConnectorDescription connector : connectors) {
  		Class<? extends Connector> clazz = connector.getConnectorClass();
  		if (isFilter(clazz)) {
  			filters.add(connector);
  		}
    }
  	return filters;
  }

  private boolean isConnector(Class<?> clazz) {
  	try {
  		Object connector = clazz.newInstance();
  		return connector instanceof Connector;
  	} catch (Exception e) {
			return false;
		}
  }

  private boolean isMapper(Class<?> clazz) {
  	try {
  		Object connector = clazz.newInstance();
  		return connector instanceof Mapper;
  	} catch (Exception e) {
			return false;
		}
  }

  private boolean isFilter(Class<?> clazz) {
  	try {
  		Object connector = clazz.newInstance();
  		return connector instanceof Filter;
  	} catch (Exception e) {
			return false;
		}
  }

  private boolean isRoleResolver(Class<?> clazz) {
  	try {
  		Object connector = clazz.newInstance();
  		return connector instanceof RoleResolver;
  	} catch (Exception e) {
			return false;
		}
  }

  public List<ConnectorDescription> getAllConnectors(String categoryId) {
  	return getConnectors(connectors, categoryId);
  }

  public List<ConnectorDescription> getJavaConnectors(String categoryId) {
  	List<ConnectorDescription> javaConnectors = getJavaConnectors();
  	return getConnectors(javaConnectors, categoryId);
  }

  public List<ConnectorDescription> getRoleResolverConnectors(String categoryId) {
  	List<ConnectorDescription> roleMappers = getRoleResolvers();
  	return getConnectors(roleMappers, categoryId);
  }

  public List<ConnectorDescription> getFilterConnectors(String categoryId) {
  	List<ConnectorDescription> mappers = getFilters();
  	return getConnectors(mappers, categoryId);
  }

  private List<ConnectorDescription> getConnectors(List<ConnectorDescription> list, String categoryId) {
    List<ConnectorDescription> connectors = new ArrayList<ConnectorDescription>();
    for (ConnectorDescription connector : list) {
      String cat = connector.getCategoryId();
      if (cat == null) {
        cat = "Other";
      }
      if (categoryId.equals(cat)) {
      	connectors.add(connector);
      }
    }
    return connectors;
  }

  public ConnectorDescription getConnector(String id) {
  	return getConnector(connectors, id);
  } 

  public ConnectorDescription getRoleResolverConnector(String id) {
  	List<ConnectorDescription> roleMappers = getRoleResolvers();
  	return getConnector(roleMappers, id);
  }

  public ConnectorDescription getJavaConnector(String id) {
  	List<ConnectorDescription> javaConnectors = getJavaConnectors();
  	return getConnector(javaConnectors, id);
  }

  public ConnectorDescription getFilterConnector(String id) {
  	List<ConnectorDescription> mappers = getFilters();
  	return getConnector(mappers, id);
  }

  private ConnectorDescription getConnector(List<ConnectorDescription> list, String id) {
  	ConnectorDescription desc = null;
    int size = list.size();
    int i = 0;
    boolean found = false;
    while (i < size && !found) {
      ConnectorDescription temp = list.get(i);
      if (id.equals(temp.getId())) {
        desc = temp;
        found = true;
      }
      i++;
    }
    if (!found) {
      throw new IllegalArgumentException(id + " is not a valid identifier");
    }
    return desc;
  }

  public List<String> getAllCategories() {
    return getCategories(connectors);
  }

  public List<String> getJavaConnectorsCategories() {
  	List<ConnectorDescription> javaConnectors = getJavaConnectors();
    return getCategories(javaConnectors);
  }

  public List<String> getRoleResolversCategories() {
    List<ConnectorDescription> roleMappers = getRoleResolvers();
    return getCategories(roleMappers);
  }

  public List<String> getFiltersCategories() {
  	List<ConnectorDescription> mappers = getFilters();
    return getCategories(mappers);
  }
  
  public String getCategoryLabel(String categoryId) {
    return getPropertyValue(categoryId);
  }

  private List<String> getCategories(List<ConnectorDescription> list) {
    List<String> categories = new ArrayList<String>();
    boolean other = false;
    for (ConnectorDescription connector : list) {
      String categoryId = connector.getCategoryId();
      if (categoryId != null) {
        String cat = getPropertyValue(categoryId);
        if (cat != null) {
          if (!categories.contains(categoryId)) {
            categories.add(categoryId);
          }
        } else {
          other = true;
        }
      } else {
        other = true;
      }
    }
    Collections.sort(categories);
    if (other) {
      categories.add("Other");
    }
    return categories;
  }

  private String getPropertyValue(String property) {
    try {
      ResourceBundle bundle =
        ResourceBundle.getBundle(this.getClass().getName(), currentLocale);
      return bundle.getString(property);
    } catch (Exception e) {
      return null;
    }
  }
  
  private boolean isAnInstanceOfConnector(Class<?> c) {
    if (Modifier.isAbstract(c.getModifiers()) || c.equals(Connector.class)) {
      return false;
    }
    return isAConnector(c);
  }
  
  private boolean isAConnector(Class<?> c) {
    if (c.equals(Connector.class)) {
      return true;
    }
    Class<?> superClass = c.getSuperclass();
    if (superClass != null) {
      return isAConnector(superClass);
    }
    return false;
  }

	public void setCurrentLocale(Locale currentLocale) {
	  this.currentLocale = currentLocale;
  }

	public Locale getCurrentLocale() {
	  return currentLocale;
  }
}
