/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 * 
 * Modifed by Matthieu Chaffotte - BonitaSoft S.A.
 **/
package org.ow2.bonita.connector.core;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.ResourceBundle;

import org.ow2.bonita.connector.core.desc.Component;
import org.ow2.bonita.connector.core.desc.ConnectorDescriptor;
import org.ow2.bonita.connector.core.desc.Page;

public class ConnectorDescription {

  private Class <? extends Connector> classConnector;
  private Locale currentLocale;
  private ConnectorDescriptor descriptor;

  public ConnectorDescription(Class<? extends Connector> classConnector) {
    this(classConnector, Locale.getDefault());
  }

  public ConnectorDescription(Class<? extends Connector> classConnector, Locale locale) {
    if (classConnector == null) {
      throw new IllegalArgumentException("The connector class cannot be null");
    } else {
      try {
        List<ConnectorError> errors = Connector.validateConnector(classConnector);
        if (!errors.isEmpty()) {
          for (ConnectorError e : errors) {
            System.out.println(e.getField() + " " + e.getError());
          }
          throw new IllegalArgumentException("The connector contains errors!" + classConnector);
        } else {
          this.classConnector = classConnector;
          this.currentLocale = locale;
          this.descriptor = Connector.load(classConnector);
        }
      } catch (Exception e) {
        e.printStackTrace();
        throw new IllegalArgumentException(classConnector.getSimpleName() + " cannot find the connector descriptor");
      } 
    }
  }

  /**
   * 
   * @param locale
   */
  public void setLocale(Locale locale) {
    currentLocale = locale;
  }

  /**
   * Gets the Connector Identification.
   * It checks whether the Connector has an identification
   * and returns either the Connector identification or the
   * Connector class name.
   * @return the connector identification or the Connector class name.
   */
  public String getId() {
    String id = descriptor.getConnectorId();
    if (id == null) {
      id = classConnector.getSimpleName();
    }
    return id;
  }
  
  public String getConnectorLabel() {
    String connector = getResourceBundleValue("ConnectorId");
    if (connector == null) {
      return getId();
    } else {
      return connector;
    }
  }

  /**
   * Gets the Connector Description.
   * It checks whether the Connector has a description and returns either
   * the Connector Description or null
   * @return the connector description or null
   */
  public String getDescription() {
    return getResourceBundleValue("Description");
  }

  /**
   * Gets the Connector Category Id.
   * It checks whether the Connector has a category and returns either
   * the Connector Category or null
   * @return the connector category or null
   */
  public String getCategoryId() {
    return descriptor.getCategoryId();
  }

  /**
   * Gets all the Page identifiers.
   * it checks whether the Connector has page names and returns either
   * an array of page names or an empty array.
   * @return an array of page identifiers or an empty array
   */
  public List<String> getPages() {
    List<String> pagesId= new ArrayList<String>();
    List<Page> pages = descriptor.getPages();
    if (pages != null) {
      for (Page page : pages) {
        pagesId.add(page.getPageId()); 
      }
    }
    return pagesId;
  }

  /**
   * Gets the page name
   * @param pageId the page identifier
   * @return the page description
   */
  public String getPageName(String pageId) {
    return getResourceBundleValue(pageId);
  }

  /**
   * Gets the page description
   * @param pageId the page identifier
   * @return the page description
   */
  public String getPageDescription(String pageId) {
    return getResourceBundleValue(pageId + ".description");
  }

  /**
   * Gets all the components of the connector inputs in a random order
   * @return the list of all the connector components
   */
  public List<Component> getAllInputs() {
    List<Page> pages = descriptor.getPages();
    List<Component> components = new ArrayList<Component>();
    if (pages != null) {
      for (Page page : pages) {
        components.addAll(page.getWidgets());
      }
    }
    return components;
  }
  
  private Page getPage(String pageId) {
    List<Page> pages = descriptor.getPages();
    for (Page page : pages) {
      if (pageId.equals(page.getPageId())) {
        return page;
      }
    }
    return null;
  }

  /**
   * Gets all components of a connector page. If the page does not exist,
   * an Exception is raised
   * @param pageId the page identifier
   * @return the list of the components of a connector page
   */
  public List<Component> getAllPageInputs(String pageId) {
    Page page = getPage(pageId);
    if (page == null) {
      throw new IllegalArgumentException(pageId + " does not refer to a Connector page Id");
    }
    return page.getWidgets();
  }

  /**
   * Gets the label value of an input widget. If the label identifier does not refer
   * to any label value, null is returned.
   * @param labelId the labelId
   * @return the i18n label value
   */
  public String getInputLabel(String labelId) {
    String label = getResourceBundleValue(labelId + ".label");
    if (label == null) {
      return labelId;
    } else {
      return label;
    }
  }

  /**
   * Gets the connector icon or null if the connector has not got an icon.
   * @return
   */
  public InputStream getIcon() {
    String icon = descriptor.getIcon();
    if (icon == null) {
      return null;
    } else {
      return classConnector.getResourceAsStream(icon);
    }
  }

  private ResourceBundle getResourceBundle() {
    String language = descriptor.getLanguage();
    if (language == null) {
      return null;
    } else {
      return ResourceBundle.getBundle(language, currentLocale);
    }
  }

  private String getResourceBundleValue(String key) {
    ResourceBundle messages = getResourceBundle();
    if (messages == null) {
      return null;
    }
    try {
      return messages.getString(key);
    } catch(Exception e) {
      return null;
    }
  }

  public Class<? extends Connector> getConnectorClass() {
    return classConnector;
  }

  @Override
  public boolean equals(Object object) {
    boolean equals = false;
    if (object != null && object instanceof ConnectorDescription) {
      ConnectorDescription description = (ConnectorDescription) object;
      if (this.getId().equals(description.getId())) {
        equals = true;
      }
    }
    return equals;
  }
}
