/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 * 
 * Modified by Matthieu Chaffotte - BonitaSoft S.A.
 **/
package org.ow2.bonita.definition.activity;

import java.io.ByteArrayInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.lang.reflect.Method;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.bonita.connector.core.Connector;
import org.ow2.bonita.connector.core.Filter;
import org.ow2.bonita.connector.core.ProcessConnector;
import org.ow2.bonita.connector.core.desc.ConnectorDescriptor;
import org.ow2.bonita.connector.core.desc.Setter;
import org.ow2.bonita.connector.core.fileproperties.ConnectorFileProperties;
import org.ow2.bonita.connector.core.fileproperties.ConnectorParameter;
import org.ow2.bonita.connector.impl.mapper.PerformerAssignFilter;
import org.ow2.bonita.definition.Hook;
import org.ow2.bonita.definition.MultiInstantiator;
import org.ow2.bonita.definition.MultiInstantiatorDescriptor;
import org.ow2.bonita.definition.PerformerAssign;
import org.ow2.bonita.definition.RoleMapper;
import org.ow2.bonita.definition.TxHook;
import org.ow2.bonita.facade.def.element.ConnectorDefinition;
import org.ow2.bonita.facade.def.element.HookDefinition;
import org.ow2.bonita.facade.def.element.HookDefinition.Event;
import org.ow2.bonita.facade.def.majorElement.ActivityDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.exception.BonitaWrapperException;
import org.ow2.bonita.facade.exception.HookInvocationException;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.uuid.ActivityInstanceUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.runtime.ClassDataLoader;
import org.ow2.bonita.runtime.InternalExecution;
import org.ow2.bonita.runtime.InternalInstance;
import org.ow2.bonita.services.Repository;
import org.ow2.bonita.util.AccessorUtil;
import org.ow2.bonita.util.BonitaException;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.bonita.util.EngineEnvTool;
import org.ow2.bonita.util.ExceptionManager;
import org.ow2.bonita.util.GroovyUtil;
import org.ow2.bonita.util.Misc;

public class ConnectorExecutorImpl implements ConnectorExecutor {

  private static final Logger LOG = Logger.getLogger(ConnectorExecutorImpl.class.getName());

  public ConnectorExecutorImpl() { }

  /* (non-Javadoc)
   * @see org.ow2.bonita.definition.activity.HookExecutor#
   * executeHook(org.ow2.bonita.runtime.internalExecution, java.lang.String, org.ow2.bonita.definition.JavaHook)
   */
  public void executeConnector(final InternalExecution execution, final String activityName, final ConnectorDefinition connector) {
    if (connector != null) {
      final InternalInstance instance = execution.getInstance();
      final ProcessInstanceUUID instanceUUID = instance.getUUID();
      final ActivityInstanceUUID activityInstanceUUID = execution.getCurrentActivityInstanceUUID();
      final ProcessDefinitionUUID processUUID = instance.getProcessDefinitionUUID();
      final boolean inTransaction  = connector.isThrowingException();
      try {
        final Object hookInstance = ClassDataLoader.getInstance(Hook.class, instance.getProcessDefinitionUUID(), connector);
        if (!inTransaction && hookInstance instanceof TxHook) {

          String message = ExceptionManager.getInstance().getFullMessage(
              "bsi_HEI_1", TxHook.class.getName(), Hook.class.getName());
          throw new BonitaRuntimeException(message);
        }
        if (LOG.isLoggable(Level.FINE)) {
          LOG.fine("Starting hook (instance=" + instanceUUID + ", process=" + processUUID
              + ", activityId=" + activityName + ") : " + connector);
        }
        final ActivityInstance activityInst =
          EngineEnvTool.getJournalQueriers().getActivityInstance(activityInstanceUUID);
        if (hookInstance instanceof Hook) {
          final Hook hook = (Hook) hookInstance;
          hook.execute(AccessorUtil.getQueryAPIAccessor(), activityInst);
        } else if (hookInstance instanceof TxHook) {
          final TxHook txHook = (TxHook) hookInstance;
          Map<String, Object[]> allparameters = getAllParameters(connector.getParameters(), txHook, instance);
          Map<String, Object[]> inputs = new HashMap<String, Object[]>();
          if (allparameters != null) {
            inputs = getInputs(allparameters);
          }
          if (txHook instanceof ProcessConnector) {
            inputs.put("setProcessInstanceUUID", new Object[]{instanceUUID});
            inputs.put("setProcessDefinitionUUID", new Object[]{processUUID});
            inputs.put("setActivityInstanceUUID", new Object[]{activityInstanceUUID});
          }
          if (!inputs.isEmpty()) {
            setParameters(inputs, txHook, activityInstanceUUID, null);
          }
          ClassLoader baseClassLoader = Thread.currentThread().getContextClassLoader();
          try {
            ClassLoader processClassLoader = ClassDataLoader.getProcessClassLoader(processUUID);
            Thread.currentThread().setContextClassLoader(processClassLoader);
            txHook.execute(AccessorUtil.getAPIAccessor(), activityInst);

            if (allparameters != null) {
              Map<String, Object[]> outputs = getOuputs(allparameters);
              for (Entry<String, Object[]> output : outputs.entrySet()) {
                String methodName = output.getKey();
                if (methodName.startsWith("get")) {
                  String variableId = (String) output.getValue()[0];
                  Method m = txHook.getClass().getMethod(methodName, new Class[0]);
                  Object variableValue = m.invoke(txHook, new Object[0]);
                  // check activity or process
                  AccessorUtil.getRuntimeAPI().setVariable(activityInstanceUUID, variableId.toString(), variableValue);
                }
              }
            }
          } finally {
            Thread.currentThread().setContextClassLoader(baseClassLoader);
          }
        } else {
          String message = ExceptionManager.getInstance().getFullMessage(
              "bsi_HEI_2", Hook.class.getName(), TxHook.class.getName());
          throw new BonitaRuntimeException(message);
        }
        if (LOG.isLoggable(Level.FINE)) {
          LOG.fine("Finished hook (instance=" + instanceUUID + ", process=" + processUUID
              + ", activityId=" + activityName + ") : " + connector);
        }
      } catch (final Exception e) {
        e.printStackTrace();
        if (inTransaction) {
          throw new BonitaWrapperException(
              new HookInvocationException("bsi_HEI_3", connector.toString(), e)
          );
        } else if (LOG.isLoggable(Level.SEVERE)) {
          LOG.severe("Exception caught while executing hook (instance=" + instanceUUID + ", process=" + processUUID
              + ", activityId=" + activityName + ") : " + connector + " - Exception : " + Misc.getStackTraceFrom(e));
        }
      }
    }
  }

  private Map<String, Object[]> getOuputs(Map<String, Object[]> connectorParameters) {
    Map<String, Object[]> outputs = new HashMap<String, Object[]>();
    for (Entry<String, Object[]> parameter : connectorParameters.entrySet()) {
      String methodName = parameter.getKey();
      if (methodName.startsWith("get") || methodName.startsWith("is")) {
        outputs.put(methodName, parameter.getValue());
      }
    }
    return outputs;
  }

  private Map<String, Object[]> getInputs(Map<String, Object[]> connectorParameters) {
    Map<String, Object[]> inputs = new HashMap<String, Object[]>();
    for (Entry<String, Object[]> parameter : connectorParameters.entrySet()) {
      String methodName = parameter.getKey();
      if (methodName.startsWith("set")) {
        inputs.put(methodName, parameter.getValue());
      }
    }
    return inputs;
  }

  private Map<String, Object[]> getAllParameters(Map<String, Object[]> connectorParameters, Object connector,
      InternalInstance instance) {
    if (connectorParameters == null) {
      return null;
    }
    Map<String, Object[]> inputs = new HashMap<String, Object[]>();
    if (containsFileParameters(connectorParameters)) {
      File file = getFileEntry(connectorParameters);
      inputs.putAll(getInputStreamParameters(file, connector, instance));
      connectorParameters.remove(file.getFile());
    }
    inputs.putAll(connectorParameters);
    return inputs;
  }

  private Setter getSetter(ConnectorDescriptor desc, String parameterName) {
    List<Setter> inputs = desc.getInputs();
    if (inputs != null) {
      for (Setter setter : inputs) {
        if (setter.getSetterName().equals(parameterName)) {
          return setter;
        }
      }
    }
    return null;
  }

  private Object convertIfPossible(Object variableValue, String dataTypeClassName) {
    try {
      return Misc.checkAssignmentCompatibility("", variableValue, dataTypeClassName);
    } catch (Exception e) {
      return variableValue;
    }
  }

  @SuppressWarnings("unchecked")
  private void setParameters(Map<String, Object[]> parameters, Object connector,
      ActivityInstanceUUID activityInstanceUUID, ProcessInstanceUUID processInstanceUUID) throws BonitaException {
    for (Entry<String, Object[]> parameter : parameters.entrySet()) {
      String methodName = parameter.getKey();
      Object[] methodParameters =
        evaluateParametersWithGroovy(parameter.getValue(), activityInstanceUUID, processInstanceUUID);
      if (connector instanceof Connector) {
        ConnectorDescriptor desc = Connector.load((Class<? extends Connector>) connector.getClass());
        Setter setter = getSetter(desc, methodName);
        if (setter != null) {
          Object[] setterParameters = setter.getParameters();
          for (int i = 0; i < setterParameters.length; i++) {
            Object setterParameter = setterParameters[i];
            Class<?> setterParameterClass = setterParameter.getClass();
            if (!setterParameterClass.equals(methodParameters[i].getClass())) {
              Class<?> methodParameterClass = methodParameters[i].getClass();
              if (methodParameterClass.equals(BigDecimal.class) || methodParameterClass.equals(BigInteger.class)) {
                methodParameters[i] = convertIfPossible(methodParameters[i].toString(), setterParameterClass.getName());
              } else if (methodParameterClass.equals(String.class)) {
                methodParameters[i] = convertIfPossible(methodParameters[i], setterParameterClass.getName());
              }
            }
          }
        }
      }

      Class<?>[] parameterClasses = new Class[methodParameters.length];
      for (int i = 0; i < parameterClasses.length; i++) {
        parameterClasses[i] = methodParameters[i].getClass();
      }

      try {
        Method m = getMethod(connector.getClass(), methodName, parameterClasses);
        if (m == null) {
          StringBuilder builder = new StringBuilder();
          if (parameterClasses != null) {
            builder.append("[");
            for (Class<?> class1 : parameterClasses) {
              builder.append(class1);
              builder.append(",");
            }
            builder.insert(builder.length() - 1, "");
            builder.append("]");
          }
          throw new BonitaRuntimeException("Unable to find a method with name: " + methodName + " and parameters: " + builder.toString() + " in connector: " + connector.getClass());
        }
        m.invoke(connector, methodParameters);
      } catch (Exception e) {
        e.printStackTrace();
        throw new BonitaRuntimeException(e.getMessage(),e.getCause());
      } 
    }
  }

  @SuppressWarnings("unchecked")
  private Object[] evaluateParametersWithGroovy(Object[] parameters, ActivityInstanceUUID activityInstanceUUID, ProcessInstanceUUID processInstanceUUID) {
    Object[] methodParameters = parameters;
    for (int i = 0; i < parameters.length; i++) {
      Object methodParameter = methodParameters[i];
      if (methodParameter instanceof List<?>) {
        List<Object> temp = (List<Object>) methodParameter;
        for (int j = 0; j < temp.size(); j++) {
          temp.set(j, getEvaluatedExpression(temp.get(j), activityInstanceUUID, processInstanceUUID));
        }
        methodParameters[i] = temp;
      } else {
        methodParameters[i] = getEvaluatedExpression(methodParameter, activityInstanceUUID, processInstanceUUID);
      }
    }
    return methodParameters;
  }

  private Method getMethod(Class<?> connectorClass, String methodName, Class<?>[] paramTypes) {
    try {
      return connectorClass.getMethod(methodName, paramTypes);
    } catch (Exception e) {
      if (paramTypes != null) {
        Method[] methods = connectorClass.getMethods();
        for (Method method : methods) {
          if (methodName.equals(method.getName())) {
            Class<?>[] types = method.getParameterTypes();
            for (int i = 0; i < types.length; i++) {
              if (!(types[i].isAssignableFrom(paramTypes[i]) || paramTypes[i].isAssignableFrom(types[i]))) {
                break;
              }
            }
            return method;
          }
        }
      }
      return null;
    }
  }

  private Object getEvaluatedExpression(Object parameterMethod, ActivityInstanceUUID activityUUID, ProcessInstanceUUID processUUID) {
    if (parameterMethod instanceof String) {
      String expression = (String) parameterMethod;
      if (GroovyUtil.isGroovyExpression(expression)) {
        try {
          if (activityUUID != null) {
            return AccessorUtil.getRuntimeAPI().evaluateGroovyExpression(expression, activityUUID);
          } else {
            return AccessorUtil.getRuntimeAPI().evaluateGroovyExpression(expression, processUUID);
          }
        } catch (Exception e) {
          throw new BonitaRuntimeException(e.getMessage(),e.getCause());
        }
      }
    }
    return parameterMethod;
  }

  private boolean containsFileParameters(Map<String, Object[]> parameters) {
    boolean inFile = false;
    if (parameters.containsKey("file:") || parameters.containsKey("bar:")
        || parameters.containsKey("resource:")) {
      inFile = true;
    }
    return inFile;
  }

  private File getFileEntry(Map<String, Object[]> parameters) {
    if (parameters.containsKey("file:")) {
      return new File("file:", (String) parameters.get("file:")[0]);
    } else if (parameters.containsKey("bar:")){
      return new File("bar:", (String) parameters.get("bar:")[0]);
    } else if (parameters.containsKey("resource:")) {
      return new File("resource:", (String) parameters.get("resource:")[0]);
    } else {
      return null;
    }
  }

  private Map<String, Object[]> getInputStreamParameters(File file, Object connector, InternalInstance instance) {
    InputStream input = getInputStream(file, instance);
    Map<String, Object[]> fileParameters = new HashMap<String, Object[]>();
    if (input != null) {
      List<ConnectorParameter> props = null;
      ConnectorFileProperties properties;
      try {
        properties = new ConnectorFileProperties(input);
        props = properties.getConnectorProperties();
      } catch (Exception e) {
        e.printStackTrace();
      }
      // Only one parameter support TODO more than one
      for (ConnectorParameter prop : props) {
        String args = prop.getValue();
        args = ((String) args).replaceAll("&amp;", "&");
        args = ((String) args).replaceAll("&lt;", "<");
        fileParameters.put(prop.getSetterName(), new Object[]{args});
      }
    }
    return fileParameters;
  }

  private InputStream getInputStream(File file, InternalInstance instance) {
    InputStream input = null;
    if ("file:".equals(file.getFile())) {
      input = getInputFileParameters(file.getPath());
    } else if ("resource:".equals(file.getFile()))  {
      input = getInputFileAsAResource(file.getPath());
    } else if ("bar:".equals(file.getFile()))  {
      ProcessDefinitionUUID processDefUUID = instance.getProcessDefinitionUUID();
      input = getInputFileParametersInBar(processDefUUID, file.getPath());
    }
    return input;
  }

  /* (non-Javadoc)
   * @see org.ow2.bonita.definition.activity.HookExecutor#
   * executeMultiInstantiator(org.ow2.bonita.runtime.internalExecution, java.lang.String, org.ow2.bonita.definition.MultiInstantiator)
   */
  public MultiInstantiatorDescriptor executeMultiInstantiator(
      final InternalExecution execution,
      final String activityId,
      final MultiInstantiator actInstantiator) throws Exception {
    return actInstantiator.execute(
        AccessorUtil.getQueryAPIAccessor(),
        execution.getInstance().getUUID(), activityId, execution.getIterationId());
  }

  public Set<String> executeFilter(final Filter filter, final PerformerAssign performerAssign,
      final ActivityInstance activityInstance, final Set<String> candidates,
      final Map<String, Object[]>parameters) throws Exception {
    Repository repository = EngineEnvTool.getRepository();
    InternalInstance instance = repository.getInstance(activityInstance.getProcessInstanceUUID());
    Map<String, Object[]> allparameters = getAllParameters(parameters, filter, instance);
    Map<String, Object[]> inputs = new HashMap<String, Object[]>();
    if (filter instanceof PerformerAssignFilter) {
      inputs.put("setPerformerAssign", new Object[]{performerAssign});
      if (parameters != null) {
        setParameters(parameters, performerAssign, activityInstance.getUUID(), null);
      }
    } else if (allparameters != null) {
      inputs = getInputs(allparameters);
    }
    inputs.put("setCandidates", new Object[]{candidates});
    inputs.put("setActivityInstanceUUID", new Object[]{activityInstance.getUUID()});
    if (!inputs.isEmpty()) {
      setParameters(inputs, filter, activityInstance.getUUID(), null);
    }
    filter.execute();
    return filter.getMembers();
  }

  /* (non-Javadoc)
   * @see org.ow2.bonita.definition.activity.HookExecutor#
   * executeRoleMapper(org.ow2.bonita.definition.RoleMapper, org.ow2.bonita.facade.uuid.ProcessInstanceUUID, java.lang.String)
   */
  public Set<String> executeRoleMapper(final RoleMapper roleMapper,
      final ProcessInstanceUUID instanceUUID,
      final String roleId, Map<String, Object[]> parameters) throws Exception {
    Repository repository = EngineEnvTool.getRepository();
    InternalInstance instance = repository.getInstance(instanceUUID);
    Map<String, Object[]> allparameters = getAllParameters(parameters, roleMapper, instance);
    if (allparameters != null) {
      Map<String, Object[]> inputs = getInputs(allparameters);
      setParameters(inputs, roleMapper, null, instanceUUID);
    }
    return roleMapper.searchMembers(AccessorUtil.getQueryAPIAccessor(), instanceUUID, roleId);
  }

  /* (non-Javadoc)
   * @see org.ow2.bonita.definition.activity.HookExecutor#
   * executeHooks(java.util.List, org.ow2.bonita.runtime.internalExecution, java.lang.String, org.ow2.bonita.definition.JavaHook.Type)
   */
  public void executeConnectors(final ActivityDefinition activityDef, final InternalExecution execution, final Event event) {
    List<HookDefinition> hooks = activityDef.getHooks(); 
    if (hooks != null) {
      for (final HookDefinition hook : hooks) {
        if (hook.getEvent() != null && hook.getEvent().equals(event)) {
          executeConnector(execution, activityDef.getName(), hook);
        }
      }
    }
  }

  private static InputStream getInputFileParameters(String variable) {
    try {
      return new FileInputStream(variable);
    } catch (FileNotFoundException e) {
      return null;
    }
  }

  private static InputStream getInputFileParametersInBar(
      ProcessDefinitionUUID processDefUUID, String variable) {
    try {
      ProcessDefinition processDef =
        AccessorUtil.getQueryDefinitionAPI().getProcess(processDefUUID);
      byte[] data = processDef.getBusinessArchive().getResource(variable);
      return new ByteArrayInputStream(data);
    } catch (Exception e) {
      e.printStackTrace();
      return null;
    }
  }

  private static InputStream getInputFileAsAResource(String variable) {
    InputStream input =
      ClassDataLoader.getCommonClassLoader().getResourceAsStream(variable);
    return input;
  }

  private class File {

    private String file;
    private String path;

    public File(String file, String path) {
      super();
      this.file = file;
      this.path = path;
    }

    public String getFile() {
      return file;
    }

    public String getPath() {
      return path;
    }
  }
}
