/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.definition.activity;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.bonita.definition.InternalProcess;
import org.ow2.bonita.facade.def.majorElement.ActivityDefinition;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.pvm.Execution;
import org.ow2.bonita.pvm.activity.ActivityExecution;
import org.ow2.bonita.runtime.InternalExecution;
import org.ow2.bonita.runtime.InternalInstance;
import org.ow2.bonita.services.Recorder;
import org.ow2.bonita.services.Repository;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.bonita.util.EngineEnvTool;
import org.ow2.bonita.util.ExceptionManager;

/**
 * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes, Pierre Vigneras
 */
public class SubFlow extends AbstractActivity {

  /**
   * 
   */
  private static final long serialVersionUID = 477565487347215726L;

  protected static final Logger LOG = Logger.getLogger(SubFlow.class.getName());

  public static final String SUBFLOW_SIGNAL = "end_of_subflow";

  protected  SubFlow() {
    super();
  }

  public SubFlow(ActivityDefinition activityDef) {
    super(activityDef);
  }

  @Override
  protected boolean bodyStartAutomatically() {
    return true;
  }

  @Override
  protected boolean executeBusinessLogic(final Execution execution) {
    final InternalExecution internalExecution = (InternalExecution) execution;

    final Repository repository = EngineEnvTool.getRepository();
    
    String subflowProcessName = this.activityDef.getSubflowProcessName();
    final InternalProcess process = repository.findLatestProcessByName(subflowProcessName);
    if (process == null) {
    	String message = ExceptionManager.getInstance().getFullMessage(
    			"be_SF_1", subflowProcessName);
      throw new BonitaRuntimeException(message);
    }
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Starting a new instance of process (as a subProcess) : " + subflowProcessName);
    }
    final Map<String, Object> parameterValues = new HashMap<String, Object>();
    Map<String, Collection<String>> inParameters = this.activityDef.getSubflowInParameters();
    if (inParameters != null) {
      for (final Map.Entry<String, Collection<String>> parameter : inParameters.entrySet()) {
        Object parentProcessVar = internalExecution.getVariable(parameter.getKey());
        for (String subProcessVarName : parameter.getValue()) {
          parameterValues.put(subProcessVarName, parentProcessVar);
        }
      }
    }
    final InternalExecution rootExecution = (InternalExecution) process.beginProcessInstance();
    rootExecution.setVariables(parameterValues);
    final InternalInstance instance = rootExecution.getInstance();

    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Started subprocess instance : " + instance);
    }
    instance.setExecutionToSignal(internalExecution);
    internalExecution.setSubflowInstance(instance);
    rootExecution.signal();
    return false;
  }

  @Override
  public void signal(final ActivityExecution execution, final String signal, final Map<String, Object> signalParameters) {
    final InternalExecution internalExecution = (InternalExecution) execution;
    internalExecution.setSubflowInstance(null);
    if (SUBFLOW_SIGNAL.equals(signal)) {
      Map<String, Collection<String>> outParameters = this.activityDef.getSubflowOutParameters();
      if (outParameters != null) {
        final Recorder recorder = EngineEnvTool.getRecorder();
        final ProcessInstanceUUID instanceUUID = internalExecution.getInstance().getUUID();
        
        for (final Map.Entry<String, Collection<String>> parameter : outParameters.entrySet()) {
          final Object variableValue = signalParameters.get(parameter.getKey());
          for (String parentProcessVarName : parameter.getValue()) {
            execution.setVariable(parentProcessVarName, variableValue);
            
            if (internalExecution.getInstance().getRootExecution().hasVariable(parentProcessVarName)) {
              recorder.recordInstanceVariableUpdated(parentProcessVarName, variableValue, instanceUUID, EngineEnvTool.getUserId());
            } else {
              recorder.recordActivityVariableUpdated(parentProcessVarName, variableValue, internalExecution.getCurrentActivityInstanceUUID(), EngineEnvTool.getUserId());
            }
          }

          
        }
      }
      super.signal(internalExecution, BODY_FINISHED, null);
    } else {
      super.signal(internalExecution, signal, signalParameters);
    }
  }

}


