package org.ow2.bonita.env.generator;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.bonita.definition.activity.ConnectorExecutorImpl;
import org.ow2.bonita.env.EnvConstants;
import org.ow2.bonita.env.EnvironmentFactory;
import org.ow2.bonita.env.XpdlEnvironmentParser;
import org.ow2.bonita.facade.AutoDetectSecurityContext;
import org.ow2.bonita.facade.BonitaSecurityContext;
import org.ow2.bonita.persistence.db.DbHistory;
import org.ow2.bonita.persistence.db.DbJournal;
import org.ow2.bonita.persistence.db.DbRepository;
import org.ow2.bonita.persistence.db.OptimizedDbHistory;
import org.ow2.bonita.persistence.log.LoggerArchiver;
import org.ow2.bonita.persistence.log.LoggerRecorder;
import org.ow2.bonita.persistence.xml.XMLHistory;
import org.ow2.bonita.services.Archiver;
import org.ow2.bonita.services.Querier;
import org.ow2.bonita.services.Recorder;
import org.ow2.bonita.services.Repository;
import org.ow2.bonita.services.handlers.FinishedInstanceHandler;
import org.ow2.bonita.services.handlers.UndeployedProcessHandler;
import org.ow2.bonita.services.handlers.impl.ArchiveFinishedInstanceHandler;
import org.ow2.bonita.services.handlers.impl.ArchiveUndeployedProcessHandler;
import org.ow2.bonita.services.handlers.impl.DeleteFinishedInstanceHandler;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.bonita.util.ExceptionManager;
import org.ow2.bonita.util.Misc;


/**
 * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes, Pierre Vigneras
 */
public abstract class EnvGenerator {

  private static final Logger LOG = Logger.getLogger(EnvGenerator.class.getName());

  private final Map<String,EnvEntry> envFactoryEntries = new TreeMap<String, EnvEntry>();
  private final Map<String,EnvEntry> envEntries = new TreeMap<String, EnvEntry>();

  public static final String HISTORY_DEFAULT_KEY = "history";
  public static final String JOURNAL_DEFAULT_KEY = "journal";

  protected List<String> mappingFiles = new ArrayList<String>();

  public static final String INDENT = "  ";

  @SuppressWarnings("unchecked")
  public EnvGenerator(final boolean defaultConfiguration, final boolean enableDbHistory, final boolean separateJournal) {

    // Hibernate configuration for bonita core engine.
    this.addHibernateConfig(
        (defaultConfiguration ? "hibernate-core-default.properties" : "hibernate-core.properties"),
        (separateJournal ? null : EnvConstants.CORE_QUERIER_SESSION_NAME),
        EnvConstants.CORE_RUNTIME_SESSION_NAME,
        EnvConstants.CORE_HB_SESSION_NAME,
        EnvConstants.CORE_HB_SESSION_FACTORY_NAME,
        EnvConstants.CORE_HB_CONF_NAME,
        "engine",
        true);

    // Hibernate configuration for history
    this.addHibernateConfig(
        (defaultConfiguration ? "hibernate-history-default.properties" : "hibernate-history.properties"),
        EnvConstants.HISTORY_QUERIER_SESSION_NAME,
        null,
        EnvConstants.HISTORY_HB_SESSION_NAME,
        EnvConstants.HISTORY_HB_SESSION_FACTORY_NAME,
        EnvConstants.HISTORY_HB_CONF_NAME,
        "db history",
        enableDbHistory);

    // Hibernate configuration for separated journal
    this.addHibernateConfig(
        (defaultConfiguration ? "hibernate-journal-default.properties" : "hibernate-journal.properties"),
        EnvConstants.JOURNAL_QUERIER_SESSION_NAME,
        null,
        EnvConstants.JOURNAL_HB_SESSION_NAME,
        EnvConstants.JOURNAL_HB_SESSION_FACTORY_NAME,
        EnvConstants.JOURNAL_HB_CONF_NAME,
        "journal (in a separate db)",
        separateJournal);

    // Command execution configuration
    String commandServiceConfigXml = "<command-service>\n";
    commandServiceConfigXml += EnvGenerator.INDENT + "<retry-interceptor />\n";
    commandServiceConfigXml += EnvGenerator.INDENT + "<environment-interceptor />\n";
    commandServiceConfigXml += EnvGenerator.INDENT + "<standard-transaction-interceptor />\n";
    commandServiceConfigXml += "</command-service>\n";
    this.setApplicationEntry("command-service",
        "Service which executes commands",
        commandServiceConfigXml,
        true);

    this.setApplicationEntry("job-executor",
        "Service which executes jobs",
        "<job-executor threads='1' auto-start='true' />",
        true);

    this.setApplicationEntry("type-resolver",
        "Mapping from java variable to database type",
        "<variable-types resource='bonita.type.resolver.xml' />",
        true);

    this.setApplicationEntry("hook-executor",
        "Implementation of the HookExecutor interface which will be used to execute hooks in the engine.",
        ConnectorExecutorImpl.class,
        true);

    this.setRepositoryType(DbRepository.class, "<arg><string value='"
        + EnvConstants.CORE_RUNTIME_SESSION_NAME + "' /></arg>");

    this.setSecurityType(AutoDetectSecurityContext.class);

    this.setBlockEntry("transaction",
        "DB transaction",
        "<transaction />",
        true);

    this.setJournalType(DbJournal.class,
        (!separateJournal ?
            "<arg><string value='" + EnvConstants.CORE_QUERIER_SESSION_NAME + "' /></arg>"
            + Misc.LINE_SEPARATOR
            + "<!-- replace previous line by this line to use a separate journal -->" + Misc.LINE_SEPARATOR
            + "<!-- <arg><string value='" + EnvConstants.JOURNAL_QUERIER_SESSION_NAME + "' /></arg> -->"
        :
          "<arg><string value='" + EnvConstants.JOURNAL_QUERIER_SESSION_NAME + "' /></arg>"));


    this.setFinishedInstanceHandlerType(DeleteFinishedInstanceHandler.class, ArchiveFinishedInstanceHandler.class);

    this.setUndeployedProcessHandlerType(ArchiveUndeployedProcessHandler.class);


    this.setBlockEntry("timer_session",
        "Service which schedules timers.",
        "<timer-session />",
        true);
    this.setBlockEntry("job-session",
        "Service which stores jobs in the database.",
        "<job-db-session session='" + EnvConstants.CORE_HB_SESSION_NAME + "' />",
        true);
    this.setBlockEntry("message_session",
        "Service which schedules messages.",
        "<message-session />",
        true);

    // recorder
    final List<EnvEntry> recorders = new ArrayList<EnvEntry>();
    recorders.add(new EnvEntry("Log recorder",
        "Recorder which writes recorded data to logs.",
        "<recorder class='" + LoggerRecorder.class.getName() + "' />",
        true));
    recorders.add(
        new EnvEntry(EnvGenerator.JOURNAL_DEFAULT_KEY,
            "Reference to the journal",
            "<ref object='" + EnvGenerator.JOURNAL_DEFAULT_KEY + "' />",
            true)
    );
    final EnvEntry recorderEnvEntry = new ChainerEnvEntry(Recorder.DEFAULT_KEY,
        "List of recorders.",
        recorders,
        true);
    this.envEntries.put(Recorder.DEFAULT_KEY, recorderEnvEntry);

    // archiver
    final List<EnvEntry> archivers = new ArrayList<EnvEntry>();
    archivers.add(new EnvEntry("Log archiver",
        "Archiver which writes archived data to logs.",
        "<archiver class='" + LoggerArchiver.class.getName() + "' />",
        true));
    archivers.add(
        new EnvEntry(EnvGenerator.HISTORY_DEFAULT_KEY,
            "Reference to the history",
            "<ref object='" + EnvGenerator.HISTORY_DEFAULT_KEY + "' />",
            true)
    );
    final EnvEntry archiverEnvEntry = new ChainerEnvEntry(Archiver.DEFAULT_KEY,
        "List of archivers.",
        archivers,
        true);
    this.envEntries.put(Archiver.DEFAULT_KEY, archiverEnvEntry);

    // Query API
    final List<EnvEntry> queryApiEntries = new ArrayList<EnvEntry>();
    queryApiEntries.add(
        new EnvEntry(EnvGenerator.JOURNAL_DEFAULT_KEY,
            "Reference to the journal",
            "<ref object='" + EnvGenerator.JOURNAL_DEFAULT_KEY + "' />",
            true)
    );
    queryApiEntries.add(
        new EnvEntry(EnvGenerator.HISTORY_DEFAULT_KEY,
            "Reference to the history",
            "<ref object='" + EnvGenerator.HISTORY_DEFAULT_KEY + "' />",
            true)
    );
    final EnvEntry queryApiEntry = new QueryApiEnvEntry(Querier.DEFAULT_KEY,
        "List of Queriers to use.",
        queryApiEntries,
        true);
    this.envEntries.put(queryApiEntry.getName(), queryApiEntry);

    final EnvEntry xmlHistoryEntry = new EnvEntry("XML history",
        "XML Implementation of the history.",
        "<history name='" + EnvGenerator.HISTORY_DEFAULT_KEY + "' class='" + XMLHistory.class.getName() + "'/>",
        !enableDbHistory);
    this.envEntries.put(xmlHistoryEntry.getName(), xmlHistoryEntry);

    final EnvEntry dbHistoryEntry = new EnvEntry("Db history",
        "DB Implementation of the history. This implementation is full conformant with the spec (but not optimized).",
        "<history name='" + EnvGenerator.HISTORY_DEFAULT_KEY + "' class='" + DbHistory.class.getName() + "'/>",
        false);
    this.envEntries.put(dbHistoryEntry.getName(), dbHistoryEntry);

    final EnvEntry optimizeDbHistoryEntry = new EnvEntry("Optimized DB history",
        "Optimized DB Implementation of the history. "
        + "This implementation contains optimized methods. "
        + "For example, a call to getUserTasks(final String userId, final ActivityState taskState) "
        + " looks in DB only if the state is an expected one in History (FINISHED is one, READY is not...)",
        "<history name='" + EnvGenerator.HISTORY_DEFAULT_KEY + "' class='" + OptimizedDbHistory.class.getName() + "'>"
        + Misc.LINE_SEPARATOR
        + EnvGenerator.INDENT + "<arg><string value='" + EnvConstants.HISTORY_QUERIER_SESSION_NAME + "' /></arg>"
        + Misc.LINE_SEPARATOR
        + "</history>",
        enableDbHistory);
    this.envEntries.put(optimizeDbHistoryEntry.getName(), optimizeDbHistoryEntry);

  }

  private void addHibernateConfig(final String propertiesName,
      final String querierSessionName,
      final String runtimeSessionName,
      final String hibernateSessionName,
      final String hibernateSessionFactoryName,
      final String hibernateConfigurationName,
      final String usage,
      final boolean enable
      ) {
    final StringBuffer hibernateConfigXml = new StringBuffer();
    hibernateConfigXml
    .append("<hibernate-configuration name='" + hibernateConfigurationName + "' >")
      .append(Misc.LINE_SEPARATOR)
      .append(EnvGenerator.INDENT).append("<properties resource='" + propertiesName + "' />")
      .append(Misc.LINE_SEPARATOR);
    if (runtimeSessionName != null) {
      hibernateConfigXml
      .append(EnvGenerator.INDENT).append("<mappings resource='bonita.runtime.mappings.hbm.xml' />")
      .append(Misc.LINE_SEPARATOR);
    }
    if (querierSessionName != null) {
      if (runtimeSessionName == null) {
        hibernateConfigXml
          .append(EnvGenerator.INDENT).append("<mappings resource='bonita.history.mappings.hbm.xml' />")
          .append(Misc.LINE_SEPARATOR);
      }
      hibernateConfigXml
      .append(EnvGenerator.INDENT).append("<mappings resource='bonita.querier.mappings.hbm.xml' />")
      .append(Misc.LINE_SEPARATOR)
      .append(EnvGenerator.INDENT).append("<cache-configuration resource='bonita.querier.cache.xml' usage='read-write' />")
      .append(Misc.LINE_SEPARATOR);
    }
    hibernateConfigXml
      .append("</hibernate-configuration>").append(Misc.LINE_SEPARATOR);

    this.setApplicationEntry(hibernateConfigurationName,
        "Configuration of the Hibernate session factory used by " + usage + ".",
        hibernateConfigXml.toString(),
        enable);


    this.setApplicationEntry(hibernateSessionFactoryName,
        "Hibernate session factory used by " + usage + ".",
        "<hibernate-session-factory name='"
        + hibernateSessionFactoryName
        + "' configuration='" + hibernateConfigurationName + "' init='eager'/>",
        enable);

    this.setBlockEntry(hibernateSessionName,
        "Hibernate session used by " + usage + ".",
        "<hibernate-session name='" + hibernateSessionName
        + "' factory='" + hibernateSessionFactoryName + "' />",
        enable);
    if (runtimeSessionName != null) {
      this.setBlockEntry(runtimeSessionName,
          "Runtime session used in " + usage + ".",
          "<" + EnvConstants.RUNTIME_DB_SESSION_TAG + " name='" + runtimeSessionName
          + "' session='" + hibernateSessionName + "'/>",
          true);
    }
    if (querierSessionName != null) {
      this.setBlockEntry(querierSessionName,
        "Querier used in " + usage + ".",
        "<" + EnvConstants.QUERIER_DB_SESSION_TAG + " name='" + querierSessionName
        + "' session='" + hibernateSessionName + "'/>",
        enable);
    }
  }

  private String getIndent(final int depth) {
    String res = "";
    for (int i = 0; i <= depth; i++) {
      res += EnvGenerator.INDENT;
    }
    return res;
  }


  /**
   * Generates a xml for the given environment configuration.
   *
   * @return
   */
  public String createEnvironmentXml() {

    int indentDepth = 0;


    // ENV XML GENERATION BEGINS HERE
    final StringBuilder result =
      new StringBuilder("<environment-definition>" + Misc.LINE_SEPARATOR + Misc.LINE_SEPARATOR);

    // DEFINITION OF THE ENVIRONMENT FACTORY
    indentDepth++;
    result.append(this.getIndent(indentDepth) + "<environment-factory>" + Misc.LINE_SEPARATOR);

    indentDepth++;
    for (final EnvEntry entry : this.envFactoryEntries.values()) {
      result.append(entry.getEnvXml(this.getIndent(indentDepth))).append(Misc.LINE_SEPARATOR);
    }
    indentDepth--;
    result
    .append(this.getIndent(indentDepth)).append("</environment-factory>" + Misc.LINE_SEPARATOR + Misc.LINE_SEPARATOR);
    // END OF ENVIRONMENT FACTORY
    // DEFINITION OF THE ENVIRONMENT

    result
    .append(this.getIndent(indentDepth)).append("<environment>" + Misc.LINE_SEPARATOR);
    indentDepth++;


    // other entries
    for (final EnvEntry entry : this.envEntries.values()) {
      result.append(entry.getEnvXml(this.getIndent(indentDepth))).append(Misc.LINE_SEPARATOR);
    }

    indentDepth--;
    result.append(this.getIndent(indentDepth)).append("</environment>" + Misc.LINE_SEPARATOR + Misc.LINE_SEPARATOR);
    indentDepth--;
    result.append(this.getIndent(indentDepth)).append("</environment-definition>" + Misc.LINE_SEPARATOR);

    return result.toString();
  }

  /**
   * Creates a new EnvironmentFactory from the list of entries.
   */
  public EnvironmentFactory createEnvironmentFactory() {
    final String envConfig = this.createEnvironmentXml();
    if (EnvGenerator.LOG.isLoggable(Level.CONFIG)) {
      EnvGenerator.LOG.config("The following environment has been generated by: "
          + this.getClass().getName() + Misc.LINE_SEPARATOR + envConfig);
    }
    final EnvironmentFactory environmentFactory = XpdlEnvironmentParser
    .parseEnvironmentFactoryFromXmlString(envConfig);
    return environmentFactory;
  }

  /**
   * Return a string that declares an object to be added in the context.
   *
   * @param classToUse
   *            classToUse at runtime
   * @return a string that declares an object to be added in the context.
   */
  @SuppressWarnings("unchecked")
  private static String getObjectDecl(final Class classToUse) {
    return "<object class='" + classToUse.getName() + "' />";
  }

  /**
   * Return a string that declares an object to be added in the context.
   *
   * @param key
   *            objectName
   * @param classToUse
   *            classToUse at runtime
   * @return a string that declares an object to be added in the context.
   */
  @SuppressWarnings("unchecked")
  private static String getObjectDecl(final String key, final Class classToUse) {
    return "<object name='" + key + "' class='" + classToUse.getName() + "' />";
  }

  /**
   * Add or replace an entry in the environment to generate.
   *
   * @param name
   *            name of the entry
   * @param xml
   *            declaration of this entry
   * @return previous declaration for this name
   */
  public void setApplicationEntry(final String name, final String description, final String xml, final boolean enabled) {
    final EnvEntry entry = new EnvEntry(name, description, xml, enabled);
    this.envFactoryEntries.put(name, entry);
  }

  /**
   * Add or replace an entry in the environment to generate.
   * 
   * @param name the name of the entry
   * @param description the entry description
   * @param classToUse implementation
   * @param enabled   is this entry enabled or disabled.
   */
  public void setApplicationEntry(final String name, final String description,
      final Class< ? > classToUse, final boolean enabled) {
    final String xml = EnvGenerator.getObjectDecl(name, classToUse);
    this.setApplicationEntry(name, description, xml, enabled);
  }
  
  /**
   * Add or replace an entry in the environment to generate.
   * 
   * @param name the name of the entry
   * @param description the entry description
   * @param xml the declaration of this entry
   * @param enabled  is this entry enabled or disabled.
   */
  public void setBlockEntry(final String name, final String description, final String xml, final boolean enabled) {
    final EnvEntry entry = new EnvEntry(name, description, xml, enabled);
    this.envEntries.put(name, entry);
  }

  /**
   * Prints the default environment on standard output.
   *
   * @param args
   * @throws IOException
   */
  @SuppressWarnings("unchecked")
  public static void main(final String[] args) throws IOException {
    if (args.length != 1 && args.length != 2) {
      System.exit(1);
    }
    EnvGenerator envGenerator = null;
    try {
      final Class <EnvGenerator> envGeneratorClass =
        (Class<EnvGenerator>) EnvGenerator.class.getClassLoader().loadClass(args[0]);
      envGenerator = envGeneratorClass.newInstance();
    } catch (final Exception e) {
      throw new RuntimeException(e);
    }
    if (args.length == 1) {
      if (EnvGenerator.LOG.isLoggable(Level.INFO)) {
        EnvGenerator.LOG.config(envGenerator.createEnvironmentXml());
      }
    } else {
      Misc.write(envGenerator.createEnvironmentXml(), new File(args[1]));
    }
  }

  public void setFinishedInstanceHandlerType(final Class< ? extends FinishedInstanceHandler>... finishedInstanceHandlerTypes) {
    final List<EnvEntry> fihEntries = new ArrayList<EnvEntry>();
    for (final Class< ? extends FinishedInstanceHandler> finishedInstanceHandlerType : finishedInstanceHandlerTypes) {
      fihEntries.add(new EnvEntry(finishedInstanceHandlerType.getName(), "", EnvGenerator.getObjectDecl(finishedInstanceHandlerType), true));
    }

    final EnvEntry entry = new ChainerEnvEntry(FinishedInstanceHandler.DEFAULT_KEY,
        "List of services called when an instance is finished.",
        fihEntries,
        true);
    this.envFactoryEntries.put(entry.getName(), entry);
  }

  public void setUndeployedProcessHandlerType(final Class< ? extends UndeployedProcessHandler>... undeployedProcessHandlerTypes) {
    final List<EnvEntry> uphEntries = new ArrayList<EnvEntry>();
    for (final Class< ? extends UndeployedProcessHandler> undeployedProcessHandlerType : undeployedProcessHandlerTypes) {
      uphEntries.add(new EnvEntry(undeployedProcessHandlerType.getName(), "", EnvGenerator.getObjectDecl(undeployedProcessHandlerType), true));
    }

    final EnvEntry entry = new ChainerEnvEntry(UndeployedProcessHandler.DEFAULT_KEY,
        "List of services called when a package is undeployed.",
        uphEntries,
        true);
    this.envFactoryEntries.put(entry.getName(), entry);
  }

  public void setJournalType(final Class< ? extends Recorder> journalType, final String... args) {
    this.checkExtendingQuerier(journalType);
    String journalXml =
      "<journal name='" + EnvGenerator.JOURNAL_DEFAULT_KEY + "' class='" + journalType.getName() + "'> " + Misc.LINE_SEPARATOR;
    for (final String arg : args) {
      journalXml += EnvGenerator.INDENT + arg + Misc.LINE_SEPARATOR;
    }
    journalXml += "</journal>";

    final EnvEntry entry = new EnvEntry(EnvGenerator.JOURNAL_DEFAULT_KEY,
        "Implementation of the journal.",
        journalXml,
        true);
    this.envEntries.put(entry.getName(), entry);
  }

  private void checkExtendingQuerier(final Class < ? > clazz) {
    if (!(Querier.class.isAssignableFrom(clazz))) {
    	String message = ExceptionManager.getInstance().getFullMessage(
    			"bse_EG_1", clazz.getName(), Querier.class);
      throw new BonitaRuntimeException(message);
    }
  }

  public void setRepositoryType(final Class< ? extends Repository> repositoryType, final String... args) {
    String repositoryXml = "<repository class='" + repositoryType.getName() + "'>" + Misc.LINE_SEPARATOR;
    for (final String arg : args) {
      repositoryXml += EnvGenerator.INDENT + arg + Misc.LINE_SEPARATOR;
    }
    repositoryXml += "</repository>";

    final EnvEntry entry = new EnvEntry("repository",
        "Implementation of the repository.",
        repositoryXml,
        true);
    this.envEntries.put(entry.getName(), entry);
  }

  public void setSecurityType(final Class< ? extends BonitaSecurityContext> securityType) {
    final EnvEntry entry = new EnvEntry("security",
        "Service used to find the user logged in.",
        "<security class='" + securityType.getName() + "'/>",
        true);
    this.envEntries.put(entry.getName(), entry);
  }

  public List<String> getMappingFiles() {
    return this.mappingFiles;
  }
}
