/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.facade.impl;

import java.io.IOException;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Map;
import java.util.Set;

import org.ow2.bonita.definition.ClassData;
import org.ow2.bonita.deployment.Deployer;
import org.ow2.bonita.deployment.DeploymentRuntimeException;
import org.ow2.bonita.facade.def.element.BusinessArchive;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessFullDefinition;
import org.ow2.bonita.facade.exception.DeploymentException;
import org.ow2.bonita.facade.exception.ProcessNotFoundException;
import org.ow2.bonita.facade.exception.UndeletableInstanceException;
import org.ow2.bonita.facade.exception.UndeletableProcessException;
import org.ow2.bonita.facade.internal.InternalManagementAPI;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.runtime.ClassDataLoader;
import org.ow2.bonita.runtime.InternalInstance;
import org.ow2.bonita.services.Archiver;
import org.ow2.bonita.services.Querier;
import org.ow2.bonita.services.Recorder;
import org.ow2.bonita.services.Repository;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.bonita.util.ClassDataTool;
import org.ow2.bonita.util.EngineEnvTool;
import org.ow2.bonita.util.ExceptionManager;
import org.ow2.bonita.util.Misc;

/**
 * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes,
 *         Pierre Vigneras
 */
public class ManagementAPIImpl implements InternalManagementAPI {

  protected ManagementAPIImpl() {
  }

  public ProcessDefinition deploy(final BusinessArchive businessArchive) throws DeploymentException {
    FacadeUtil.checkArgsNotNull(businessArchive);
    try {
      return Deployer.deploy(businessArchive);
    } catch (final Exception e) {
      throw new DeploymentException(e.getMessage(), e);
    }
  }

  public void deployClass(final byte[] clazz) throws DeploymentException {
    FacadeUtil.checkArgsNotNull(clazz);
    String className;
    try {
      className = ClassDataTool.visitClass(clazz).getClassName();
    } catch (final ArrayIndexOutOfBoundsException e) {
      String message = ExceptionManager.getInstance().getFullMessage("bai_MAPII_4");
      throw new DeploymentException(message);
    }
    final Repository repository = EngineEnvTool.getRepository();
    final ClassData globalClassData = repository.getGlobalClassData(className);
    if (globalClassData != null) {
      // this class was already deployed
      String message = ExceptionManager.getInstance().getFullMessage("bai_MAPII_5");
      throw new DeploymentException(message, className);
    }
    repository.storeGlobalClassData(new ClassData(className, clazz));
  }

  public void deployClassesInJar(final byte[] classesArchive) throws DeploymentException {
    FacadeUtil.checkArgsNotNull(classesArchive);
    try {
      Map<String, byte[]> classResources = Misc.getResourcesFromZip(classesArchive);
      final Collection<byte[]> classes = new ArrayList<byte[]>();
      for (final byte[] classData : classResources.values()) {
        classes.add(classData);
      }
      this.deployClasses(classes);
    } catch (final IOException e) {
      throw new DeploymentException(e.getMessage(), e);
    }
  }

  public void deployClasses(final Collection<byte[]> classes) throws DeploymentException {
    FacadeUtil.checkArgsNotNull(classes);
    for (final byte[] data : classes) {
      this.deployClass(data);
    }
  }

  public void undeploy(final ProcessDefinitionUUID processUUID) throws DeploymentException {
    FacadeUtil.checkArgsNotNull(processUUID);
    try {
      Deployer.undeployProcess(processUUID, EngineEnvTool.getUserId());
    } catch (final DeploymentRuntimeException e) {
      throw new DeploymentException(e.getMessage(), e);
    }
  }

  public void removeClass(final String className) throws DeploymentException {
    FacadeUtil.checkArgsNotNull(className);

    final Repository repository = EngineEnvTool.getRepository();
    if (repository.getGlobalClassData(className) == null) {
      String message = ExceptionManager.getInstance().getFullMessage("bai_MAPII_6");
      throw new DeploymentException(message, className);
    }
    // check that no process is still using this class
    //for all process having a dependency on this class, if the class is not available at process level, we can't remove it
    final Set<ProcessDefinitionUUID> dependentProcesses = repository.getDependentProcesses(className);
    if (dependentProcesses != null && !dependentProcesses.isEmpty()) {
      for (final ProcessDefinitionUUID processUUID : dependentProcesses) {
        final Collection<ClassData> datas = repository.getProcessClassDatas(processUUID);
        if (datas != null) {
          boolean classExistAtProcessLevel = false;
          for (ClassData classData : datas) {
            if (classData.getClassName().equals(className)) {
              classExistAtProcessLevel = true;
              break;
            }
          }
          if (!classExistAtProcessLevel) {
            //the class to remove is not in the process class datas.
            //It means, it is a common class and it is used by a process: can't remove it.
            String message = ExceptionManager.getInstance().getFullMessage("bai_MAPII_7");
            throw new DeploymentException(message,
                className, processUUID);
          }
        }
      }
    }
    repository.removeGlobalClassData(className);
    ClassDataLoader.removeCommonClass(className);
  }

  public void removeClasses(final String[] classNames) throws DeploymentException {
    if (classNames != null) {
      for (final String className : classNames) {
        this.removeClass(className);
      }
    }
  }

  public void replaceClass(final String className, final byte[] newClazz) throws DeploymentException {
    this.removeClass(className);
    this.deployClass(newClazz);
  }

  public void deleteProcess(final ProcessDefinitionUUID processUUID) throws ProcessNotFoundException, 
  UndeletableProcessException, UndeletableInstanceException {
    FacadeUtil.checkArgsNotNull(processUUID);

    final Querier journal = EngineEnvTool.getJournalQueriers();
    final Querier history = EngineEnvTool.getHistoryQueriers();
    final Repository repository = EngineEnvTool.getRepository();

    ProcessFullDefinition processDef = journal.getProcess(processUUID);
    boolean inJournal = processDef != null;
    if (processDef == null) {
      processDef = history.getProcess(processUUID);
    }

    if (processDef == null) {
      throw new ProcessNotFoundException("bai_MAPII_9", processUUID);
    }

    final RuntimeAPIImpl runtimeAPI = new RuntimeAPIImpl();
    try {
      runtimeAPI.deleteAllProcessInstances(processUUID);
      repository.removeProcessVersion(processDef.getName(), processDef.getVersion());
    } catch (final ProcessNotFoundException e) {
      throw new BonitaRuntimeException(e);
    }

    if (inJournal) {
      final Set<InternalInstance> instances = repository.getInstances(processUUID);
      if (!instances.isEmpty()) {
        final ProcessInstanceUUID instanceUUID = instances.iterator().next().getUUID();
        throw new UndeletableProcessException("bai_MAPII_10", processUUID, instanceUUID);
      }
      Deployer.removeProcessDependencies(processDef);
      final Recorder recorder = EngineEnvTool.getRecorder();
      recorder.remove(processDef);
    }

    if (!inJournal) {
      final Archiver archiver = EngineEnvTool.getArchiver();
      archiver.remove(processDef);
    }

  }

  public void deleteAllProcesses() throws UndeletableInstanceException,
  RemoteException, UndeletableProcessException {
    Querier querier = EngineEnvTool.getAllQueriers();
    for (ProcessDefinition processDefinition : querier.getProcesses()) {
      try {
        deleteProcess(processDefinition.getUUID());
      } catch (ProcessNotFoundException e) {
        //nothing
      }
    }
  }

  public String getLoggedUser() throws RemoteException {
    return EngineEnvTool.getUserId();
  }

  public void addMetaData(String key, String value) throws RemoteException {
    EngineEnvTool.getRepository().storeMetaData(key, value);
  }

  public void deleteMetaData(String key) throws RemoteException {
    EngineEnvTool.getRepository().deleteMetaData(key);
  }

  public String getMetaData(String key) throws RemoteException {
    return EngineEnvTool.getRepository().getMetaData(key);
  }
}
