/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 * 
 * Modified by Matthieu Chaffotte - BonitaSoft S.A.
 **/
package org.ow2.bonita.function;

import java.util.List;

import org.ow2.bonita.connector.core.ConnectorError;

/**
 * The <code>SubStringStringConnector</code> class wraps both methods
 * substring of the <code>String</code> class. The connector can use
 * two indexes: the first one to start the cut and the second one to finish
 * the cut.The connector uses the substring method with the simple parameter
 * when endIndex is <code>null</code> otherwise it uses the method with both
 * indexes.
 * @see java.lang.String#substring(int)
 * @see java.lang.String#substring(int, int)
 * @author chaffotm
 *
 */
public class SubStringStringConnector extends StringConnector {

  private Integer beginIndex;
  private Integer endIndex;

  public Integer getBeginIndex() {
    return beginIndex;
  }

  public Integer getEndIndex() {
    return endIndex;
  }

  public void setBeginIndex(int beginIndex) {
    this.beginIndex = beginIndex;
  }

  /**
   * Converts and sets the ending index. If beginIndex is greater than
   * Integer.MAX_VALUE or less than Integer.MIN_VALUE, beginIndex equals
   * -1.
   * @param beginIndex the beginning index
   */
  public void setBeginIndex(Long beginIndex) {
    if (beginIndex == null) {
      this.beginIndex = null;
    } else {
      this.beginIndex = beginIndex.intValue();
    }
  }

  public void setEndIndex(int endIndex) {
    this.endIndex = endIndex;
  }

  /**
   * Converts and sets the ending index. If endIndex is greater than
   * Integer.MAX_VALUE or less than Integer.MIN_VALUE, endIndex equals
   * -1.
   * @param endIndex the ending index
   */
  public void setEndIndex(Long endIndex) {
    if (endIndex == null) {
      this.endIndex = null;
    } else {
      this.endIndex = endIndex.intValue();
    }
  }

  @Override
  protected void executeConnector() throws Exception {
  	String temp = getString();
    if (getEndIndex() == null) {
      temp = temp.substring(beginIndex);
    } else {
      temp = temp.substring(beginIndex, endIndex);
    }
    setResult(temp);
  }

  @Override
  protected List<ConnectorError> validateValues() {
    List<ConnectorError> errors = super.validateValues();
    if (getBeginIndex() < 0) {
      errors.add(new ConnectorError("beginIndex",
          new IllegalArgumentException("cannot be negative")));
    } else if (getBeginIndex() > getString().length()) {
      errors.add(new ConnectorError("beginIndex",
          new IllegalArgumentException(
              "cannot be greater than the String lenght!")));
    }
    if (getEndIndex() != null) {
      if (getEndIndex() < 0) {
        errors.add(new ConnectorError("endIndex",
            new IllegalArgumentException("cannot be negative")));
      } else {
        if (getEndIndex() > getString().length()) {
          errors.add(new ConnectorError("endIndex",
              new IllegalArgumentException(
                  "cannot be greater than the String lenght!")));
        }
        if (getBeginIndex() > getEndIndex()) {
          errors.add(new ConnectorError("beginIndex",
              new IllegalArgumentException(
                  "cannot be greater than the endIndex!")));
        }
      }
    }
    return errors;
  }
}
