/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 * 
 * Mofified by Charles Souillard - BonitaSoft S.A.
 **/
package org.ow2.bonita.persistence.db;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.ow2.bonita.facade.def.majorElement.ActivityDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessFullDefinition;
import org.ow2.bonita.facade.runtime.ActivityFullInstance;
import org.ow2.bonita.facade.runtime.ActivityState;
import org.ow2.bonita.facade.runtime.ProcessFullInstance;
import org.ow2.bonita.facade.uuid.ActivityDefinitionUUID;
import org.ow2.bonita.facade.uuid.ActivityInstanceUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.persistence.QuerierDbSession;
import org.ow2.bonita.persistence.QuerierUtil;
import org.ow2.bonita.services.Querier;
import org.ow2.bonita.util.EnvTool;
import org.ow2.bonita.util.Misc;


public class AbstractDbQuerier implements Querier {

	private String persistenceServiceName;
	protected DbQuerierBuffer buffer;

	public AbstractDbQuerier(String persistenceServiceName, DbQuerierBuffer buffer) {
		super();
		this.persistenceServiceName = persistenceServiceName;
		this.buffer = buffer;
	}

	protected QuerierDbSession getQuerierDbSession() {
		return EnvTool.getQuerierDbSession(persistenceServiceName);
	}


	public ActivityDefinition getActivity(ActivityDefinitionUUID activityDefinitionUUID) {
		Misc.checkArgsNotNull(activityDefinitionUUID);
		for (ProcessFullDefinition process : buffer.getProcesses().values()) {

			Set<ActivityDefinition> activities = process.getActivities();
			if (activities != null) {
				for (ActivityDefinition activity : activities) {
					if (activity.getUUID().equals(activityDefinitionUUID)) {
						return activity;
					}
				}
			}
		}

		ActivityDefinition dbAD = getQuerierDbSession().getActivityDefinition(activityDefinitionUUID);
		if (dbAD == null || buffer.isProcessRemoved(dbAD.getProcessDefinitionUUID())) {
			return null;
		}
		return dbAD;
	}

	public ActivityFullInstance getActivityInstance(
			ProcessInstanceUUID instanceUUID, String activityName, String iterationId,
			String activityInstanceId) {

		Misc.checkArgsNotNull(instanceUUID, activityName, iterationId, activityInstanceId);

		if (buffer.isInstanceRemoved(instanceUUID)) {
			return null;
		}

		ProcessFullInstance processFullInstance = buffer.getInstance(instanceUUID);
		if (processFullInstance != null) {
			return processFullInstance.getActivity(activityName, iterationId, activityInstanceId);
		}
		return getQuerierDbSession().getActivityInstance(instanceUUID, activityName, iterationId, activityInstanceId);
	}

	public ActivityFullInstance getActivityInstance(
			ActivityInstanceUUID activityInstanceUUID) {

		Misc.checkArgsNotNull(activityInstanceUUID);
		for (ProcessFullInstance instance : buffer.getInstances().values()) {
			Set<ActivityFullInstance> activities = instance.getActivities();
			if (activities != null) {
				for (ActivityFullInstance activity : activities) {
					if (activity.getUUID().equals(activityInstanceUUID)) {
						return activity;
					}
				}
			}
		}
		ActivityFullInstance aFI = 
			getQuerierDbSession().getActivityInstance(activityInstanceUUID);
		if (aFI == null || buffer.isInstanceRemoved(aFI.getProcessInstanceUUID())) {
			return null;
		}
		return aFI;
	}

	public Set<ActivityFullInstance> getActivityInstances(
			ProcessInstanceUUID instanceUUID) {

		Misc.checkArgsNotNull(instanceUUID);

		if (buffer.isInstanceRemoved(instanceUUID)) {
			return new HashSet<ActivityFullInstance>();
		}
		Set<ActivityFullInstance> activityInstances = null;
		ProcessFullInstance processFullInstance = buffer.getInstance(instanceUUID);
		if (processFullInstance != null) {
			return processFullInstance.getActivities();
		}

		activityInstances = getQuerierDbSession().getActivityInstances(instanceUUID);
		if (activityInstances != null) {
			return activityInstances;
		}
		return new HashSet<ActivityFullInstance>();
	}


	public ProcessFullInstance getProcessInstance(ProcessInstanceUUID instanceUUID) {
		Misc.checkArgsNotNull(instanceUUID);

		if (buffer.isInstanceRemoved(instanceUUID)) {
			return null;
		}
		ProcessFullInstance processFullInstance = buffer.getInstance(instanceUUID);
		if (processFullInstance != null) {
			return processFullInstance;
		}
		processFullInstance = getQuerierDbSession().getProcessInstance(instanceUUID);
		if (processFullInstance != null) {
			buffer.addInstance(processFullInstance);
		}
		return processFullInstance;
	}

	public Set<ProcessFullInstance> getProcessInstances() {
		Set<ProcessFullInstance> dbInstances = getQuerierDbSession().getProcessInstances();
		for (ProcessFullInstance instance : dbInstances) {
			if (!buffer.containsInstance(instance.getUUID()) && !buffer.isInstanceRemoved(instance.getUUID())) {
				buffer.addInstance(instance);
			}
		}
		return buffer.getInstancesSet();
	}
	
	public Set<ProcessFullInstance> getProcessInstances(
			Collection<ProcessInstanceUUID> instanceUUIDs) {
		Set<ProcessFullInstance> dbInstances = getQuerierDbSession().getProcessInstances(instanceUUIDs);
		for (ProcessFullInstance instance : dbInstances) {
			if (!buffer.containsInstance(instance.getUUID()) && !buffer.isInstanceRemoved(instance.getUUID())) {
				buffer.addInstance(instance);
			}
		}
		return buffer.getInstancesSet();
	}
	

  
  public Set<ProcessFullInstance> getProcessInstancesWithTaskState(Collection<ActivityState> activityStates) {
    Set<ProcessFullInstance> dbInstances = getQuerierDbSession().getProcessInstancesWithTaskState(activityStates);
    for (ProcessFullInstance instance : dbInstances) {
      if (!buffer.containsInstance(instance.getUUID()) && !buffer.isInstanceRemoved(instance.getUUID())) {
        buffer.addInstance(instance);
      }
    }
    return buffer.getInstancesSet();
  }

	public Set<ProcessFullInstance> getProcessInstances(
			ProcessDefinitionUUID processUUID) {
		Misc.checkArgsNotNull(processUUID);

		Set<ProcessFullInstance> dbInstances = getQuerierDbSession().getProcessInstances(processUUID);
		for (ProcessFullInstance instance : dbInstances) {
			if (!buffer.containsInstance(instance.getUUID()) && !buffer.isInstanceRemoved(instance.getUUID())) {
				buffer.addInstance(instance);
			}
		}

		Set<ProcessFullInstance> instances = new HashSet<ProcessFullInstance>();
		for (ProcessFullInstance instance : buffer.getInstances().values()) {
			if (instance.getProcessDefinitionUUID().equals(processUUID)) {
				instances.add(instance);
			}
		}
		return instances;
	}

	public ProcessFullDefinition getProcess(String processId, String version) {
		Misc.checkArgsNotNull(processId, version);

		for (ProcessFullDefinition process : buffer.getProcesses(processId).values()) {
			if (process.getVersion().equals(version)) {
				return process;
			}
		}

		ProcessFullDefinition process = getQuerierDbSession().getProcess(processId, version);
		if (process == null || buffer.isProcessRemoved(process.getUUID())) {
			return null;
		}
		return process;
	}

	public ProcessFullDefinition getProcess(ProcessDefinitionUUID processUUID) {
		Misc.checkArgsNotNull(processUUID);

		for (ProcessFullDefinition process : buffer.getProcesses().values()) {
			if (process.getUUID().equals(processUUID)) {
				return process;
			}
		}

		ProcessFullDefinition process = getQuerierDbSession().getProcess(processUUID);
		if (process == null || buffer.isProcessRemoved(process.getUUID())) {
			return null;
		}
		return process;

	}

	public Set<ProcessFullDefinition> getProcesses() {
		Set<ProcessFullDefinition> processes = getQuerierDbSession().getProcesses();
		return mergeBuffer(processes);
	}

	public Set<ProcessFullDefinition> getProcesses(String processId) {
		Misc.checkArgsNotNull(processId);

		Set<ProcessFullDefinition> processes = getQuerierDbSession().getProcesses(processId);
		return mergeBuffer(processes);
	}

	public Set<ProcessFullDefinition> getProcesses(ProcessDefinition.ProcessState processState) {
		Misc.checkArgsNotNull(processState);

		Set<ProcessFullDefinition> processes = getQuerierDbSession().getProcesses(processState);
		return mergeBuffer(processes);
	}

	public Set<ProcessFullDefinition> getProcesses(String processId, ProcessDefinition.ProcessState processState) {
		Misc.checkArgsNotNull(processId, processState);

		Set<ProcessFullDefinition> processes = getQuerierDbSession().getProcesses(processId, processState);
		return mergeBuffer(processes);
	}

	public ActivityFullInstance getTaskInstance(ActivityInstanceUUID taskUUID) {
		Misc.checkArgsNotNull(taskUUID);

		for (ProcessFullInstance instance : buffer.getInstances().values()) {
			Set<ActivityFullInstance> activities = instance.getActivities();
			if (activities != null) {
				for (ActivityFullInstance activity : activities) {
					if (activity.getUUID().equals(taskUUID) && activity.isTask()) {
						return (ActivityFullInstance) activity;
					}
				}
			}
		}
		ActivityFullInstance aFI = 
			getQuerierDbSession().getTaskInstance(taskUUID);
		if (aFI == null || buffer.isInstanceRemoved(aFI.getProcessInstanceUUID())) {
			return null;
		}
		return aFI;

	}

	public Set<ActivityFullInstance> getTaskInstances(
			ProcessInstanceUUID instanceUUID) {
		Misc.checkArgsNotNull(instanceUUID);

		if (buffer.isInstanceRemoved(instanceUUID)) {
			return new HashSet<ActivityFullInstance>();
		}

		Set<ActivityFullInstance>taskActivities = new HashSet<ActivityFullInstance>();

		ProcessFullInstance instance = buffer.getInstance(instanceUUID);
		if (instance != null) {
			Set<ActivityFullInstance> activities = instance.getActivities();
			if (activities != null) {
				for (ActivityFullInstance activity : activities) {
					if (activity.isTask()) {
						taskActivities.add((ActivityFullInstance) activity);
					}
				}
			}
			return taskActivities;
		}
		return getQuerierDbSession().getTaskInstances(instanceUUID);
	}

	public Set<ActivityFullInstance> getUserInstanceTasks(
			String userId, ProcessInstanceUUID instanceUUID, ActivityState taskState) {
		Misc.checkArgsNotNull(userId, instanceUUID, taskState);

		if (buffer.isInstanceRemoved(instanceUUID)) {
			return new HashSet<ActivityFullInstance>();
		}

		Set<ActivityFullInstance> taskActivities = new HashSet<ActivityFullInstance>();

		ProcessFullInstance instance = buffer.getInstance(instanceUUID);
		if (instance != null) {
			taskActivities.addAll(QuerierUtil.getUserTasks(instance.getActivities(), userId, taskState));
			return taskActivities;
		}

		return getQuerierDbSession().getUserInstanceTasks(userId, instanceUUID, taskState);
	}

	public Set<ActivityFullInstance> getUserTasks(String userId,
			ActivityState taskState) {
		Misc.checkArgsNotNull(userId, taskState);

		Set<ActivityFullInstance> bufferTaskActivities = 
			new HashSet<ActivityFullInstance>();
		for (ProcessFullInstance instance : buffer.getInstances().values()) {
			bufferTaskActivities.addAll(QuerierUtil.getUserTasks(instance.getActivities(), userId, taskState));
		}

		Map<ActivityInstanceUUID, ActivityFullInstance> taskActivities = 
			new HashMap<ActivityInstanceUUID, ActivityFullInstance>();

		for (ActivityFullInstance taskActivity : bufferTaskActivities) {
			taskActivities.put(taskActivity.getUUID(), taskActivity);
		}

		Set<ActivityFullInstance> dbTaskActivities = getQuerierDbSession().getUserTasks(userId, taskState);
		if (dbTaskActivities != null) {
			for (ActivityFullInstance activity : dbTaskActivities) {
				if (!buffer.isInstanceRemoved(activity.getProcessInstanceUUID()) && !taskActivities.containsKey(activity.getUUID())) {
					taskActivities.put(activity.getUUID(), activity);
				}
			}
		}

		Set<ActivityFullInstance> result =  new HashSet<ActivityFullInstance>();
		result.addAll(taskActivities.values());
		return result;
	}

	/*
	 *SPECIFIC
	 */

	public Set<ProcessFullDefinition> getDependentProcesses(ProcessDefinitionUUID processUUID) {
		Misc.checkArgsNotNull(processUUID);

		Set<ProcessDefinitionUUID> uuidDeps = new HashSet<ProcessDefinitionUUID>();
		Set<ProcessFullDefinition> deps = getQuerierDbSession().getDependentProcesses(processUUID);
		if (deps != null) {
			for (ProcessFullDefinition process : deps) {
				uuidDeps.add(process.getUUID());
				if (buffer.isProcessRemoved(process.getUUID())) {
					deps.remove(process);
				}
			}
		}

		if (deps == null) {
			deps = new HashSet<ProcessFullDefinition>();
		}
		for (ProcessFullDefinition proc : buffer.getProcesses().values()) {
			boolean isDep = !proc.getUUID().equals(processUUID)
			&& proc.getProcessDependencies() != null && proc.getProcessDependencies().contains(processUUID.getProcessName());
			if (isDep && !uuidDeps.contains(proc.getUUID())) {
				uuidDeps.add(proc.getUUID());
				deps.add(proc);

			}
		}

		return deps;
	}

	public ProcessFullDefinition getLastDeployedProcess(String processId, ProcessDefinition.ProcessState processState) {
		Misc.checkArgsNotNull(processId, processState);

		ProcessFullDefinition lastProcess = getQuerierDbSession().getLastDeployedProcess(processId, processState);

		for (ProcessFullDefinition process : buffer.getProcesses().values()) {
			boolean match = false;
			if (process.getName().equals(processId)) {
				if (ProcessDefinition.ProcessState.DEPLOYED.equals(processState) && process.getUndeployedBy() == null) {
					match = true;
				} else if (ProcessDefinition.ProcessState.UNDEPLOYED.equals(processState)) {
					match = true;
				}
			}

			if (match && (lastProcess == null || process.getDeployedDate().after(lastProcess.getDeployedDate()))) {
				lastProcess = process;
			}
		}

		return lastProcess;
	}

	private Set<ProcessFullDefinition> mergeBuffer(Set<ProcessFullDefinition> processes) {
		if (processes == null) {
			return new HashSet<ProcessFullDefinition>();
		}
		for (ProcessFullDefinition process : processes) {
			if (buffer.isProcessRemoved(process.getUUID())) {
				processes.remove(process);
			}
		}
		return processes;
	}
}


