package org.ow2.bonita.persistence.db;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import org.ow2.bonita.definition.ClassData;
import org.ow2.bonita.definition.InternalProcess;
import org.ow2.bonita.facade.def.element.MetaData;
import org.ow2.bonita.facade.def.element.impl.MetaDataImpl;
import org.ow2.bonita.facade.uuid.ActivityInstanceUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.persistence.RuntimeDbSession;
import org.ow2.bonita.pvm.Execution;
import org.ow2.bonita.pvm.job.Timer;
import org.ow2.bonita.runtime.ActivityInstanceKey;
import org.ow2.bonita.runtime.InternalExecution;
import org.ow2.bonita.runtime.InternalInstance;
import org.ow2.bonita.runtime.ProcessInstanceKey;
import org.ow2.bonita.runtime.ProcessVersion;
import org.ow2.bonita.services.Repository;
import org.ow2.bonita.util.EnvTool;
import org.ow2.bonita.util.Misc;

/**
 * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes,
 *         Pierre Vigneras
 */
public class DbRepository implements Repository {

  private final String persistenceServiceName;
  private final DbRepositoryBuffer buffer = new DbRepositoryBuffer();

  private RuntimeDbSession getRuntimeDbSession() {
    return EnvTool.getRuntimeDbSession(this.persistenceServiceName);
  }

  public DbRepository(final String persistenceServiceName) {
    this.persistenceServiceName = persistenceServiceName;
  }

  // INSTANCE REPOSITORY
  public InternalInstance getInstance(final ProcessInstanceUUID instanceUUID) {
    Misc.checkArgsNotNull(instanceUUID);
    if (this.buffer.isInstanceRemoved(instanceUUID)) {
      return null;
    }
    InternalInstance instance = this.buffer.getInstance(instanceUUID);
    if (instance != null) {
      return instance;
    }
    instance = getRuntimeDbSession().getInstance(instanceUUID);
    if (instance != null) {
      this.buffer.addInstance(instance);
    }
    return instance;
  }

  public Set<InternalInstance> getInstances(
      final ProcessDefinitionUUID processUUID) {
    Misc.checkArgsNotNull(processUUID);
    final Set<InternalInstance> dbInstances = getRuntimeDbSession()
        .getInstances(processUUID);
    for (final InternalInstance instance : dbInstances) {
      if (!this.buffer.containsInstance(instance.getUUID())
          && !this.buffer.isInstanceRemoved(instance.getUUID())) {
        this.buffer.addInstance(instance);
      }
    }
    final Set<InternalInstance> result = new HashSet<InternalInstance>();
    for (final InternalInstance instance : this.buffer.getInstances().values()) {
      if (instance.getProcessDefinitionUUID().equals(processUUID)) {
        result.add(instance);
      }
    }
    return result;
  }

  public void storeInstance(final InternalInstance instance) {
    Misc.checkArgsNotNull(instance);
    getRuntimeDbSession().save(instance);
    this.buffer.addInstance(instance);
  }

  public InternalInstance removeInstance(final ProcessInstanceUUID instanceUUID) {
    Misc.checkArgsNotNull(instanceUUID);
    final InternalInstance instance = getInstance(instanceUUID);
    if (instance == null) {
      return null;
    }
    this.buffer.removeInstance(instance);
    getRuntimeDbSession().delete(instance);
    return instance;
  }

  private InternalExecution getExecOnNode(final InternalExecution exec,
      final ActivityInstanceUUID activityInstanceUUID) {
    Misc.checkArgsNotNull(exec, activityInstanceUUID);
    if (exec.getExecutions() == null || exec.getExecutions().isEmpty()) {
      if (exec.getNode() != null && exec.getCurrentActivityInstanceUUID() != null
          && exec.getCurrentActivityInstanceUUID().equals(activityInstanceUUID)) {
        return exec;
      }
    } else {
      for (final Execution child : exec.getExecutions()) {
        final InternalExecution found = getExecOnNode(
            (InternalExecution) child, activityInstanceUUID);
        if (found != null) {
          return found;
        }
      }
    }
    return null;
  }

  public InternalExecution getExecutionOnActivity(
      final ProcessInstanceUUID instanceUUID,
      final ActivityInstanceUUID activityUUID) {
    Misc.checkArgsNotNull(instanceUUID, activityUUID);

    final InternalInstance instance = getInstance(instanceUUID);
    if (instance != null) {
      return getExecOnNode(instance.getRootExecution(), activityUUID);
    }
    final InternalExecution exec = getRuntimeDbSession()
        .getExecutionPointingOnNode(activityUUID);
    if (exec != null
        && !this.buffer.isInstanceRemoved(exec.getInstance().getUUID())) {
      return exec;
    }
    return null;
  }

  // PROCESS REPOSITORY
  public InternalProcess getProcess(final ProcessDefinitionUUID processUUID) {
    Misc.checkArgsNotNull(processUUID);

    if (this.buffer.isProcessRemoved(processUUID)) {
      return null;
    }
    InternalProcess process = this.buffer.getProcess(processUUID);
    if (process != null) {
      return process;
    }
    process = getRuntimeDbSession().getProcess(processUUID);
    if (process != null) {
      this.buffer.addProcess(process);
    }
    return process;
  }

  public void storeProcess(final InternalProcess process) {
    Misc.checkArgsNotNull(process);
    this.buffer.addProcess(process);
    getRuntimeDbSession().save(process);
  }

  public String getLastProcessVersion(final String processId) {
    Misc.checkArgsNotNull(processId);
    final ProcessVersion pv = getRuntimeDbSession().getLastProcessVersion(
        processId);
    if (pv != null) {
      return pv.getProcessVersion();
    }
    return "";
  }

  public void storeProcessVersion(final String processId,
      final String processVersion) {
    Misc.checkArgsNotNull(processId, processVersion);
    ProcessVersion pv = getRuntimeDbSession().getLastProcessVersion(processId);
    if (pv == null) {
      pv = new ProcessVersion(processId, processVersion);
      getRuntimeDbSession().save(pv);
    } else {
      pv.setProcessVersion(processVersion);
    }
  }

  public void removeProcessVersion(final String processId,
      final String processVersion) {
    Misc.checkArgsNotNull(processId, processVersion);
    final ProcessVersion pv = getRuntimeDbSession().getLastProcessVersion(
        processId);
    if (pv != null) {
      final boolean stillExist = pv.removeVersion(processVersion);
      if (!stillExist) {
        getRuntimeDbSession().delete(pv);
      }
    }
  }

  public InternalProcess removeProcess(final ProcessDefinitionUUID processUUID) {
    final InternalProcess p = getProcess(processUUID);
    if (p == null) {
      return null;
    }
    getRuntimeDbSession().delete(p);
    return p;
  }


  // PACKAGE CLASS DATA
  public Collection<ClassData> getProcessClassDatas(
      final ProcessDefinitionUUID processUUID) {
    Misc.checkArgsNotNull(processUUID);

    if (this.buffer.isProcessClassDataRemoved(processUUID)) {
      return null;
    }
    Collection<ClassData> classDatas = this.buffer
        .getProcessClassData(processUUID);
    if (classDatas != null) {
      return classDatas;
    }
    classDatas = getRuntimeDbSession().getProcessClassDatas(processUUID);
    if (classDatas != null) {
      this.buffer.addProcessClassDatas(processUUID, classDatas);
    }
    return classDatas;
  }

  public void removeProcessClassDatas(final ProcessDefinitionUUID processUUID) {
    Misc.checkArgsNotNull(processUUID);
    final Collection<ClassData> classDatas = getProcessClassDatas(processUUID);
    if (classDatas == null) {
      return;
    }
    this.buffer.removeProcessClassDatas(processUUID);
    for (ClassData classData : classDatas) {
      getRuntimeDbSession().delete(classData);
    }
  }

  public void storeProcessClassDatas(final ProcessDefinitionUUID processUUID,
      final Collection<ClassData> classDatas) {
    Misc.checkArgsNotNull(classDatas);
    for (ClassData classData : classDatas) {
      getRuntimeDbSession().save(classData);
    }
    this.buffer.addProcessClassDatas(processUUID, classDatas);
  }

  // GLOBAL CLASS DATAS
  public ClassData getGlobalClassData(final String className) {
    Misc.checkArgsNotNull(className);
    if (this.buffer.isGlobalClassDataRemoved(className)) {
      return null;
    }

    ClassData globalCD = this.buffer.getGlobalClassData(className);
    if (globalCD != null) {
      return globalCD;
    }
    globalCD = getRuntimeDbSession().getGlobalClassData(className);
    if (globalCD != null) {
      this.buffer.addGlobalClassData(globalCD);
    }
    return globalCD;
  }

  public void removeGlobalClassData(final String className) {
    Misc.checkArgsNotNull(className);
    final ClassData gcd = getGlobalClassData(className);
    if (gcd == null) {
      return;
    }
    this.buffer.removeGlobalClassData(gcd);
    getRuntimeDbSession().delete(gcd);
  }

  public void storeGlobalClassData(final ClassData globalClassData) {
    Misc.checkArgsNotNull(globalClassData);
    getRuntimeDbSession().save(globalClassData);
    this.buffer.addGlobalClassData(globalClassData);
  }

  // DEPENDENT PACKAGES
  public Set<ProcessDefinitionUUID> getDependentProcesses(final String className) {
    Misc.checkArgsNotNull(className);
    return getRuntimeDbSession().getDependentProcesses(className);
  }

  /*
   * public void addProcessDependencies(final ProcessDefinitionUUID processUUID,
   * final Set<String> classNames) { Misc.checkArgsNotNull(processUUID);
   * getRuntimeDbSession().save(new ProcessDependency(processUUID, classNames));
   * } public ProcessDependency getProcessDependency(final ProcessDefinitionUUID
   * processUUID) { Misc.checkArgsNotNull(processUUID); return
   * getRuntimeDbSession().getProcessDependency(processUUID); } public void
   * removeProcessDependencies(final ProcessDefinitionUUID processUUID) {
   * Misc.checkArgsNotNull(processUUID); final ProcessDependency
   * processDependency = getProcessDependency(processUUID); if
   * (processDependency == null) { return; }
   * getRuntimeDbSession().delete(processDependency); }
   */
  public InternalProcess findLatestProcessByName(final String subProcessId) {
    return getRuntimeDbSession().findLatestProcessById(subProcessId);
  }

  public Set<Timer> getInstanceTimers(final ProcessInstanceUUID instanceUUID) {
    return getRuntimeDbSession().getInstanceTimers(instanceUUID);
  }

  public void removeTimer(final Timer timer) {
    EnvTool.getTimerSession().cancel(timer);
  }

  public synchronized long getNextActivityInstanceNb(
      final ProcessInstanceUUID instanceUUID, final String activityName) {
    final ActivityInstanceKey nextPik = new ActivityInstanceKey(instanceUUID,
        activityName);
    getRuntimeDbSession().save(nextPik);
    getRuntimeDbSession().delete(nextPik);
    return nextPik.getNumber();
  }

  public synchronized long getNextProcessInstanceNb(
      final ProcessDefinitionUUID processUUID) {
    final ProcessInstanceKey nextPik = new ProcessInstanceKey(processUUID);
    getRuntimeDbSession().save(nextPik);
    getRuntimeDbSession().delete(nextPik);
    return nextPik.getNumber();
  }

  public void storeMetaData(String key, String value) {
    Misc.checkArgsNotNull(key, value);
    MetaDataImpl data = (MetaDataImpl) getRuntimeDbSession().getMetaData(key);
    if (data == null) {
      data = new MetaDataImpl(key, value);
    } else {
      data.setValue(value);
    }
    getRuntimeDbSession().save(data);
  }

  public String getMetaData(String key) {
    Misc.checkArgsNotNull(key);
    MetaData data = getRuntimeDbSession().getMetaData(key);
    if (data == null) {
      return null;
    } else {
      return data.getValue();
    }
  }

  public void deleteMetaData(String key) {
    Misc.checkArgsNotNull(key);
    MetaData data = getRuntimeDbSession().getMetaData(key);
    getRuntimeDbSession().delete(data);
  }
}
