/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.persistence.log;

import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.bonita.facade.def.majorElement.ProcessFullDefinition;
import org.ow2.bonita.facade.runtime.ProcessFullInstance;
import org.ow2.bonita.facade.uuid.ActivityInstanceUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.services.Recorder;
import org.ow2.bonita.services.info.ActivityInstanceCurrentInfo;
import org.ow2.bonita.services.info.ProcessInstanceCurrentInfo;
import org.ow2.bonita.services.info.TaskCurrentInfo;

/**
 * @author Pierre Vigneras
 * @date Jan 31, 2008
 */
public class LoggerRecorder implements Recorder {
  private static final Logger LOG = Logger.getLogger(LoggerRecorder.class.getName());
  private final Level level;

  public LoggerRecorder() {
    this(Level.FINE);
  }

  public LoggerRecorder(final String levelName) {
    this(Level.parse(levelName));
  }

  public LoggerRecorder(final Level level) {
    this.level = level;
  }

  public void recordProcessDeployed(final ProcessFullDefinition processDef) {
    if (LOG.isLoggable(this.level)) {
      LOG.log(this.level, "Deployed procss: processDefinitionUUID: " + processDef.getUUID()
          + ", processId: " + processDef.getName() + ", userId: "
          + processDef.getDeployedBy() + ", description: " + processDef.getDescription());
    }
  }

  public void recordProcessUndeployed(final ProcessFullDefinition processDef, final String userId) {
    if (LOG.isLoggable(this.level)) {
      LOG.log(this.level, "Undeployed process: processDefinitionUUID: "
          + processDef.getUUID() + ", userId = " + userId);
    }
  }

  public void recordInstanceVariableUpdated(final String variableId, final Object variableValue,
      final ProcessInstanceUUID instanceUUID, final String userId) {
    if (LOG.isLoggable(this.level)) {
      LOG.log(this.level, "Global variable updated: processUUID: "
          + instanceUUID + ", variableId: " + variableId + ", userId = " + userId);
    }
  }

  public void recordActivityVariableUpdated(final String variableId,
      final Object variableValue,
      final ActivityInstanceUUID activityUUID,
      final String userId) {
    if (LOG.isLoggable(this.level)) {
      LOG.log(this.level, "Local variable updated: activityUUID: "
          + activityUUID + ", variableId: " + variableId + ", userId = " + userId);
    }
  }

  public void remove(final ProcessFullDefinition processDef) {
    if (LOG.isLoggable(this.level)) {
      LOG.log(this.level, "ProcessDefinition " + processDef.getUUID() + " removed.");
    }
  }
  public void remove(final ProcessFullInstance processInst) {
    if (LOG.isLoggable(this.level)) {
      LOG.log(this.level, "ProcessInstance " + processInst.getUUID() + " removed.");
    }
  }
  public void clear() {
    if (LOG.isLoggable(this.level)) {
      LOG.log(this.level, "Clearing recorder");
    }
  }


  public void recordBodyEnded(final ActivityInstanceCurrentInfo activityInstanceCurrentInfo) {
    if (LOG.isLoggable(this.level)) {
      LOG.log(this.level, "Activity body ended: "
          + activityInstanceCurrentInfo.getActivityLabel()
          + " (" + activityInstanceCurrentInfo.getInstanceUUID() + ")");
    }
  }

  public void recordBodyAborted(final ActivityInstanceCurrentInfo activityInstanceCurrentInfo) {
    if (LOG.isLoggable(this.level)) {
      LOG.log(this.level, "Activity body aborted: "
          + activityInstanceCurrentInfo.getActivityLabel()
          + " (" + activityInstanceCurrentInfo.getInstanceUUID() + ")");
    }
  }

  public void recordBodyCancelled(final ActivityInstanceCurrentInfo activityInstanceCurrentInfo) {
    if (LOG.isLoggable(this.level)) {
      LOG.log(this.level, "Activity body cancelled: "
          + activityInstanceCurrentInfo.getActivityLabel()
          + " (" + activityInstanceCurrentInfo.getInstanceUUID() + ")");
    }
  }

  public void recordBodyStarted(final ActivityInstanceCurrentInfo activityInstanceCurrentInfo) {
    if (LOG.isLoggable(this.level)) {
      LOG.log(this.level, "Activity body started: "
          + activityInstanceCurrentInfo.getActivityLabel()
          + " (" + activityInstanceCurrentInfo.getInstanceUUID() + ")");
    }
  }

  public void recordEnterActivity(final ActivityInstanceCurrentInfo activityInstanceCurrentInfo) {
    if (LOG.isLoggable(this.level)) {
      LOG.log(this.level, "Activity ready: "
          + activityInstanceCurrentInfo.getActivityName()
          + " (" + activityInstanceCurrentInfo.getInstanceUUID() + ")");
    }

  }

  public void recordInstanceEnded(final ProcessInstanceCurrentInfo processInstanceCurrentInfo) {
    if (LOG.isLoggable(this.level)) {
      LOG.log(this.level, "Ended instance: "
          + processInstanceCurrentInfo.getInstanceUUID()
          + ", userId = " + processInstanceCurrentInfo.getUserId());
    }
  }

  public void recordInstanceAborted(final ProcessInstanceCurrentInfo processInstanceCurrentInfo) {
    if (LOG.isLoggable(this.level)) {
      LOG.log(this.level, "Aborted instance: "
          + processInstanceCurrentInfo.getInstanceUUID()
          + ", userId = " + processInstanceCurrentInfo.getUserId());
    }
  }

  public void recordInstanceCancelled(final ProcessInstanceCurrentInfo processInstanceCurrentInfo) {
    if (LOG.isLoggable(this.level)) {
      LOG.log(this.level, "Cancelled instance: "
          + processInstanceCurrentInfo.getInstanceUUID()
          + ", userId = " + processInstanceCurrentInfo.getUserId());
    }
  }

  public void recordInstanceStarted(final ProcessInstanceCurrentInfo processInstanceCurrentInfo) {
    if (LOG.isLoggable(this.level)) {
      LOG.log(this.level, "Started instance: "
          + processInstanceCurrentInfo.getInstanceUUID()
          + ", userId = " + processInstanceCurrentInfo.getUserId());
    }
  }


  public void recordTaskAssigned(final TaskCurrentInfo taskCurrentInfo) {
    if (LOG.isLoggable(this.level)) {
      LOG.log(this.level, "Task " + taskCurrentInfo.getTaskUUID()
          + " Assigned by: " + taskCurrentInfo.getLoggedInUserId()
          + " Assigned to: " + taskCurrentInfo.getAssignedUserId()
          + "and candidates: " + taskCurrentInfo.getCandidates());
    }
  }


  public void recordTaskFinished(final TaskCurrentInfo taskCurrentInfo) {
    if (LOG.isLoggable(this.level)) {
      LOG.log(this.level, "Task " + taskCurrentInfo.getTaskUUID() + " finished by " + taskCurrentInfo.getLoggedInUserId());
    }
  }
  public void recordTaskReady(final TaskCurrentInfo taskCurrentInfo) {
    if (LOG.isLoggable(this.level)) {
      LOG.log(this.level, "Task ready " + taskCurrentInfo.getTaskUUID());
    }
  }

  public void recordTaskResumed(final TaskCurrentInfo taskCurrentInfo) {
    if (LOG.isLoggable(this.level)) {
      LOG.log(this.level, "Task " + taskCurrentInfo.getTaskUUID()
          + " resumed to " + taskCurrentInfo.getTaskState()
          + " by " + taskCurrentInfo.getLoggedInUserId() + ".");
    }
  }

  public void recordTaskStarted(final TaskCurrentInfo taskCurrentInfo) {
    if (LOG.isLoggable(this.level)) {
      LOG.log(this.level, "Task " + taskCurrentInfo.getTaskUUID() + " started by " + taskCurrentInfo.getLoggedInUserId() + ".");
    }
  }

  public void recordTaskSuspended(final TaskCurrentInfo taskCurrentInfo) {
    if (LOG.isLoggable(this.level)) {
      LOG.log(this.level, "Task " + taskCurrentInfo.getTaskUUID() + " suspended by " + taskCurrentInfo.getLoggedInUserId() + ".");
    }
  }
}
