/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 * 
 * Modified by Charles Souillard - BonitaSoft S.A.
 * Modified by Matthieu Chaffotte - BonitaSoft S.A.
 **/
package org.ow2.bonita.persistence.xml;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.Writer;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.bonita.facade.def.majorElement.ActivityDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessFullDefinition;
import org.ow2.bonita.facade.def.majorElement.impl.ProcessDefinitionImpl;
import org.ow2.bonita.facade.runtime.ActivityFullInstance;
import org.ow2.bonita.facade.runtime.ActivityState;
import org.ow2.bonita.facade.runtime.ProcessFullInstance;
import org.ow2.bonita.facade.runtime.impl.ProcessFullInstanceImpl;
import org.ow2.bonita.facade.uuid.ActivityDefinitionUUID;
import org.ow2.bonita.facade.uuid.ActivityInstanceUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.persistence.QuerierUtil;
import org.ow2.bonita.services.Archivable;
import org.ow2.bonita.services.Archiver;
import org.ow2.bonita.services.Querier;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.bonita.util.ExceptionManager;
import org.ow2.bonita.util.Misc;

import com.thoughtworks.xstream.XStream;

/**
 *
 *
 * @author Pierre Vigneras
 * @date Feb 29, 2008
 */
public class XMLHistory implements Archiver, Querier {

	private static final String XML_EXTENSION = ".xml";
	private static final String INSTANCE_FILENAME_PREFIX = "instance-";
	private static final String PROCESS_FILENAME_PREFIX = "process-";

	private static final String DEFINITION_FOLDER = "definitions";
	private static final String RUNTIME_FOLDER = "runtime";
	private static final Logger LOG = Logger.getLogger(XMLHistory.class.getName());

	private static final Object CONSTRUCTOR_MUTEX = new Object();
	private static final String ENCODING = "UTF8";	

	private void logWrite(final Archivable archivable, final File file) {
		if (LOG.isLoggable(Level.FINE)) {
			LOG.fine("Writing: " + archivable + " to file: " + file);
		}
		Writer writer = null;
		try {
			writer = new OutputStreamWriter(new FileOutputStream(file), ENCODING);
			writer.append(xstream.toXML(archivable));
			writer.flush();
		} catch (IOException e) {
			throw new BonitaRuntimeException(e);
		} finally {
			Misc.close(writer);
		}
	}

	private Object logRead(final Class< ? extends Archivable> clazz, final File file) {
		if (LOG.isLoggable(Level.FINE)) {
			LOG.fine("Fetching class: " + clazz.getName() + " from file: " + file);
		}
		Reader reader = null;
		try {
			reader = new InputStreamReader(new FileInputStream(file), ENCODING);
			return xstream.fromXML(reader);
		} catch (IOException e) {
			throw new BonitaRuntimeException(e);
		} finally {
			Misc.close(reader);
		}
	}

	private final File base;
	private final XStream xstream;

	public XMLHistory() {
		this(new File(Misc.TMP_DIR, "Bonita"));
	}

	public XMLHistory(final String path) {
		this(new File(path));
	}

	public XMLHistory(final File path) {
		this.base = new File(path, "history");
		if (LOG.isLoggable(Level.CONFIG)) {
			LOG.config("Configuring history: " + XMLHistory.class.getName() + " with base: " + base);
		}
		// if history directory does not exist, create it in mutual exclusion
		// all is done in synchronized blocks, because double-checked locking pattern is broken
		synchronized (CONSTRUCTOR_MUTEX) {
      if (this.base.exists()) {
        if (!this.base.isDirectory() || !this.base.canRead() || !this.base.canWrite()) {
        	String message = ExceptionManager.getInstance().getFullMessage(
        			"bsi_XMLH_1", path);
          throw new BonitaRuntimeException(message);
        }
      } else if (!this.base.mkdirs()) {
      	String message = ExceptionManager.getInstance().getFullMessage(
      			"bsi_XMLH_2", path);
        throw new BonitaRuntimeException(message);
      }
    }
		this.xstream = new XStream();
	}

	/*
	 * PROCESS INSTANCE
	 */

	public void archive(ProcessFullInstance processInst) {
		ProcessFullInstance newInstance = new ProcessFullInstanceImpl(processInst);
		//newInstance.end();
		final File file = getFile(processInst.getProcessDefinitionUUID(),
				processInst.getProcessInstanceUUID(), INSTANCE_FILENAME_PREFIX);
		checkDuplicate(newInstance, file);
		checkWritable(file);
		logWrite(newInstance, file);
	}

	public void remove(ProcessFullInstance processInst) {
		final File file = getFile(processInst.getProcessDefinitionUUID(),
				processInst.getProcessInstanceUUID(), INSTANCE_FILENAME_PREFIX);
		deleteFile(file);
	}

	// points to: base/RUNTIME_FOLDER/processDefinitionUUID/INSTANCE_FILENAME_PREFIX+processUUID.xml
	private File getFile(final ProcessDefinitionUUID processUUID, final ProcessInstanceUUID instanceUUID, final String prefix) {
		Misc.checkArgsNotNull(processUUID, instanceUUID);
		final File runtimeFolderBase = new File(base, RUNTIME_FOLDER);
		File processFolder = new File(runtimeFolderBase, processUUID.toString());
		final File instanceFile = new File(processFolder, prefix + instanceUUID.toString() + XML_EXTENSION);
		return instanceFile;
	}

	public ProcessFullInstance getProcessInstance(ProcessInstanceUUID instanceUUID) {
		Misc.checkArgsNotNull(instanceUUID);
		final File runtimeFolder = new File(base, RUNTIME_FOLDER);
		final File[] processFolders = runtimeFolder.listFiles();
		if (processFolders == null) {
			return null;
		}
		for (File processFolder : processFolders) {
			File[] instanceFiles = processFolder.listFiles();
			if (instanceFiles != null) {
				for (File instanceFile : instanceFiles) {
					if (instanceFile.getName().equals(INSTANCE_FILENAME_PREFIX + instanceUUID.toString() + ".xml")) {
						return (ProcessFullInstance) logRead(ProcessFullInstance.class, instanceFile);
					}
				}
			}
		}
		return null;
	}
	
	public Set<ProcessFullInstance> getProcessInstances(Collection<ProcessInstanceUUID> instanceUUIDs) {
		final Set<ProcessFullInstance> result = new HashSet<ProcessFullInstance>(); 
		for (ProcessInstanceUUID instanceUUID : instanceUUIDs) {
			ProcessFullInstance instance = getProcessInstance(instanceUUID);
			if (instance != null) {
				result.add(instance);
			}
		}
		return result;
	}
	
	public Set<ProcessFullInstance> getProcessInstancesWithTaskState(Collection<ActivityState> activityStates) {
	  final Set<ProcessFullInstance> result = new HashSet<ProcessFullInstance>(); 
	  Set<ProcessFullInstance> instances = getAllInstances();
    if (instances != null) {
      for (ProcessFullInstance instance : instances) {
        Set<ActivityFullInstance> activities = getActivityInstances(instance.getProcessInstanceUUID());
        if (activities != null) {
          actLoop:for (ActivityFullInstance activity : activities) {
            if (activity.isTask()  && activityStates.contains(activity.getState())) {
              result.add(instance);
              break actLoop;
            }
          }
        }
      }
    }
    return result;
	}

	public Set<ProcessFullInstance> getProcessInstances() {
		return getAllInstances();
	}

	public Set<ProcessFullInstance> getProcessInstances(ProcessDefinitionUUID processUUID) {
		Misc.checkArgsNotNull(processUUID);
		final Set<ProcessFullInstance> result = new HashSet<ProcessFullInstance>();
		final File runtimeFolder = new File(base, RUNTIME_FOLDER);
		File processFolder = new File(runtimeFolder, processUUID.toString());
		if (processFolder.exists()) {
			File[] instanceFiles = processFolder.listFiles();
			if (instanceFiles != null) {
				for (File instanceFile : instanceFiles) {
					result.add((ProcessFullInstance) logRead(ProcessFullInstance.class, instanceFile));
				}
			}
		}
		return result;
	}

	public ActivityFullInstance getTaskInstance(ActivityInstanceUUID taskUUID) {
		Misc.checkArgsNotNull(taskUUID);
		Set<ProcessFullInstance> instances = getAllInstances();
		if (instances != null) {
			for (ProcessFullInstance instance : instances) {
				Set<ActivityFullInstance> activities = getActivityInstances(instance.getProcessInstanceUUID());
				if (activities != null) {
					for (ActivityFullInstance activity : activities) {
						if (activity.isTask() && activity.getUUID().equals(taskUUID)) {
							return (ActivityFullInstance) activity;
						}
					}
				}
			}
		}
		return null;
	}

	private Set<ProcessFullInstance> getAllInstances() {
		Set<ProcessFullInstance> instances = new HashSet<ProcessFullInstance>();
		final File runtimeFolder = new File(base, RUNTIME_FOLDER);
		final File[] processFolders = runtimeFolder.listFiles();
		if (processFolders == null) {
			return null;
		}
		for (File processFolder : processFolders) {
			for (File instanceFile : processFolder.listFiles()) {
				instances.add((ProcessFullInstance) logRead(ProcessFullInstance.class, instanceFile));
			}
		}
		return instances;
	}

	public File getBasePath() {
		return base;
	}


	private void checkWritable(final File file) {
		final File parent = file.getParentFile();
		if (!parent.exists() && !parent.mkdirs()) {
			// This test can fail because:
			// - mkdir failed because of security exception, or disk full ...
			// in this case an exception should be thrown
			// - the directory has been created after calling exists() test and before mkdir() has been called.
			// in this situation, there is no failure.
			// To solve this, we can add locks in this method,
			// or we can just check again that parent does not exist.
			// We choose the last solution.
			// TODO: see if the first call to exists() can be removed
			if (!parent.exists()) {
				String message = ExceptionManager.getInstance().getFullMessage(
		      			"bsi_XMLH_3", parent);
		        throw new BonitaRuntimeException(message);
			}
		}
	}

	private void checkDuplicate(final Archivable archivable, final File file) {
		if (file.exists()) {
			String message = ExceptionManager.getInstance().getFullMessage(
	    			"bsi_XMLH_4", archivable, file);
	      throw new IllegalArgumentException(message);
		}
	}

	public void clear() {
		if (LOG.isLoggable(Level.CONFIG)) {
			LOG.config("Clearing history: " + this + ". Deleting recursively directory: " + base);
		}
		Misc.deleteDir(getBasePath());
	}

	/*
	 * DEFINITION
	 */

	public void archive(ProcessFullDefinition processDef) {
		processDef = new ProcessDefinitionImpl(processDef);
		final File file = getFile(processDef.getUUID());
		checkWritable(file);
		logWrite(processDef, file);
	}

	public void remove(ProcessFullDefinition processDef) {
		final File file = getFile(processDef.getUUID());
		deleteFile(file);
	}

	private void deleteFile(File f) {
		boolean deleted = false;
		int retry = 5;
		while (!deleted && retry > 0)  {
			deleted = f.delete();
			retry--;
		}
		if (!deleted) {
String message = ExceptionManager.getInstance().getFullMessage(
    			"bsi_XMLH_5", f, this.getClass());
      throw new RuntimeException(message);
		}
	}
	//points to: base/DEFINITION_FOLDER/PROCESS_FILENAME_PREFIX+processUUID.xml
	private File getFile(ProcessDefinitionUUID processUUID) {
		final File definitionFolder = new File(base, DEFINITION_FOLDER);
		final File processFile = new File(definitionFolder, PROCESS_FILENAME_PREFIX + processUUID + XML_EXTENSION);
		return processFile;
	}

	public ActivityFullInstance getActivityInstance(
			ActivityInstanceUUID activityInstanceUUID) {

		Misc.checkArgsNotNull(activityInstanceUUID);
		Set<ProcessFullInstance> instances = getAllInstances();
		if (instances != null) {
			for (ProcessFullInstance instance : instances) {
				Set<ActivityFullInstance> activities = instance.getActivities();
				if (activities != null) {
					for (ActivityFullInstance activity : activities) {
						if (activity.getUUID().equals(activityInstanceUUID)) {
							return activity;
						}
					}
				}
			}
		}
		return null;
	}

	public Set<ActivityFullInstance> getUserInstanceTasks(
			String userId, ProcessInstanceUUID instanceUUID, ActivityState taskState) {
		Misc.checkArgsNotNull(instanceUUID, taskState);
		ProcessFullInstance instance = getProcessInstance(instanceUUID);
		Set<ActivityFullInstance> result = new HashSet<ActivityFullInstance>();

		if (instance == null) {
			return result;
		}

		result.addAll(QuerierUtil.getUserTasks(instance.getActivities(), userId, taskState));

		return result;
	}



	public Set<ActivityFullInstance> getUserTasks(String userId, ActivityState taskState) {
		Set<ActivityFullInstance> result = new HashSet<ActivityFullInstance>();
		Set<ProcessFullInstance> instances = getAllInstances();
		if (instances != null) {
			for (ProcessFullInstance instance : instances)   {
				Set<ActivityFullInstance> activities = instance.getActivities();
				if (activities != null) {
					Set<ActivityFullInstance> taskActivities = QuerierUtil.getUserTasks(activities, userId, taskState);
					if (taskActivities != null) {
						result.addAll(taskActivities);
					}
				}
			}
		}
		return result;
	}

	public ActivityFullInstance getActivityInstance(ProcessInstanceUUID instanceUUID,
			String activityId, String iterationId, String activityInstanceId) {
		Misc.checkArgsNotNull(instanceUUID, activityId, iterationId, activityInstanceId);
		ProcessFullInstance instance = getProcessInstance(instanceUUID);
		if (instance == null) {
			return null;
		}
		return instance.getActivity(activityId, iterationId, activityInstanceId);
	}

	public Set<ActivityFullInstance> getActivityInstances(ProcessInstanceUUID instanceUUID) {
		Misc.checkArgsNotNull(instanceUUID);
		ProcessFullInstance instance = getProcessInstance(instanceUUID);
		if (instance == null) {
			return new HashSet<ActivityFullInstance>();
		}
		Set<ActivityFullInstance> result = instance.getActivities();
		if (result == null) {
			return new HashSet<ActivityFullInstance>();
		}
		return result;
	}

	public Set<ActivityFullInstance> getTaskInstances(ProcessInstanceUUID instanceUUID) {
		Misc.checkArgsNotNull(instanceUUID);
		final Set<ActivityFullInstance> result = new HashSet<ActivityFullInstance>();

		Set<ActivityFullInstance> activities = getActivityInstances(instanceUUID);
		if (activities != null) {
			for (ActivityFullInstance activity : activities) {
				if (activity.isTask()) {
					result.add((ActivityFullInstance) activity);
				}
			}
		}
		return result;
	}

	public Set<ProcessFullDefinition> getProcesses() {
		final Set<ProcessFullDefinition> result = new HashSet<ProcessFullDefinition>();
		final File definitionFolder = new File(base, DEFINITION_FOLDER);
		final File[] processFiles = definitionFolder.listFiles();
		if (processFiles == null) {
			return result;
		}
		for (File processFile : processFiles) {
			result.add((ProcessFullDefinition) logRead(ProcessFullDefinition.class, processFile));
		}
		return result;
	}

	public Set<ProcessFullDefinition> getProcesses(String processId) {
		Misc.checkArgsNotNull(processId);
		final Set<ProcessFullDefinition> result = new HashSet<ProcessFullDefinition>();
		Set<ProcessFullDefinition> processes = getProcesses();
		if (processes != null && !processes.isEmpty()) {
			for (ProcessFullDefinition process : processes) {
				if (process.getName().equals(processId)) {
					result.add(process);
				}
			}

		}
		return result;
	}

	public Set<ProcessFullDefinition> getProcesses(String processId, String version) {
		final Set<ProcessFullDefinition> result = new HashSet<ProcessFullDefinition>();
		Set<ProcessFullDefinition> processes = getProcesses();
		if (processes != null && !processes.isEmpty()) {
			for (ProcessFullDefinition process : processes) {
				if (process.getName().equals(processId) && process.getVersion().equals(version)) {
					result.add(process);
				}
			}
		}
		return result;
	}

	public Set<ProcessFullDefinition> getProcesses(ProcessDefinition.ProcessState processState) {
		Set<ProcessFullDefinition> result = new HashSet<ProcessFullDefinition>();
		for (ProcessFullDefinition process : getProcesses()) {
			boolean match = false;
			if (ProcessDefinition.ProcessState.DEPLOYED.equals(processState) && process.getUndeployedBy() == null) {
				match = true;
			} else if (ProcessDefinition.ProcessState.UNDEPLOYED.equals(processState)) {
				match = true;
			}
			if (match) {
				result.add(process);
			}
		}
		return result;
	}

	public Set<ProcessFullDefinition> getProcesses(String processId, ProcessDefinition.ProcessState processState) {
		Set<ProcessFullDefinition> result = new HashSet<ProcessFullDefinition>();
		for (ProcessFullDefinition process : getProcesses()) {
			boolean match = false;
			if (ProcessDefinition.ProcessState.DEPLOYED.equals(processState) && process.getUndeployedBy() == null) {
				match = true;
			} else if (ProcessDefinition.ProcessState.UNDEPLOYED.equals(processState)) {
				match = true;
			}
			if (match && process.getName().equals(processId)) {
				result.add(process);
			}
		}
		return result;
	}

	public ProcessFullDefinition getProcess(ProcessDefinitionUUID processUUID) {
		Misc.checkArgsNotNull(processUUID);
		File file = getFile(processUUID);
		if (!file.exists()) {
			return null;
		}
		return (ProcessFullDefinition) logRead(ProcessFullDefinition.class, file);
	}

	public ProcessFullDefinition getProcess(String processId, String version) {
		for (ProcessFullDefinition process : getProcesses(processId)) {
			if (process.getVersion().equals(version)) {
				return process;
			}
		}
		return null;
	}

	public ProcessFullDefinition getLastDeployedProcess(String processId, ProcessDefinition.ProcessState processState) {
		Misc.checkArgsNotNull(processId);
		ProcessFullDefinition lastProcess = null;
		for (ProcessFullDefinition process : getProcesses(processId, processState)) {
			if (lastProcess == null) {
				lastProcess = process;
			} else if (process.getDeployedDate().after(lastProcess.getDeployedDate())) {
				lastProcess = process;

			}
		}
		return lastProcess;
	}

	public Set<ProcessFullDefinition> getDependentProcesses(ProcessDefinitionUUID processUUID) {
		Set<ProcessFullDefinition> result = new HashSet<ProcessFullDefinition>();
		Set<ProcessFullDefinition> deployedProcesses = getProcesses(ProcessDefinition.ProcessState.DEPLOYED);
		for (ProcessFullDefinition process : deployedProcesses) {
			if (!process.getUUID().equals(processUUID)) {
				if (process.getProcessDependencies() != null && process.getProcessDependencies().contains(processUUID.getProcessName())) {
					result.add(process);
				}
			}
		}
		return result;
	}

	public ActivityDefinition getActivity(ActivityDefinitionUUID activityDefinitionUUID) {
		Set<ProcessFullDefinition> processes = getProcesses();
		if (processes != null) {
			for (ProcessFullDefinition process : processes) {
				Set<ActivityDefinition> activities = process.getActivities();
				if (activities != null) {
					for (ActivityDefinition activity : activities) {
						if (activity.getUUID().equals(activityDefinitionUUID)) {
							return activity;
						}
					}
				}
			}
		}
		return null;
	}
}
