/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.pvm.internal.jobexecutor;

import org.ow2.bonita.env.Transaction;
import org.ow2.bonita.pvm.job.Timer;
import org.ow2.bonita.pvm.session.TimerSession;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.bonita.util.ExceptionManager;
import org.ow2.bonita.util.Log;

/**
 * Timers created with this service are committed at the end of the transaction,
 * so their execution will be late if the delay is shorter than the transaction.
 * In that case, they will be executed at the end of the transaction.
 * 
 * @author Tom Baeyens, Pascal Verdage
 */
public class JobExecutorTimerSession implements TimerSession {

  private static final Log log = Log.getLog(TimerSession.class.getName());

  /* injected */
  Transaction transaction;

  /* injected */
  JobExecutor jobExecutor;

  /* injected */
  JobDbSession jobDbSession;

  boolean jobExecutorNotificationScheduled = false;

  public void schedule(Timer timer) {
    log.debug("scheduling " + timer);
    validate(timer);
    jobDbSession.save(timer);
    if (!jobExecutorNotificationScheduled) {
      jobExecutorNotificationScheduled = true;
      transaction
          .registerSynchronization(new JobAddedNotification(jobExecutor));
    }
  }

  protected void validate(Timer timer) {
    if (timer == null) {
    	String message = ExceptionManager.getInstance().getFullMessage("bp_JETS_1");
    	throw new BonitaRuntimeException(message);
    }
    if (timer.getExecution() == null) {
    	String message = ExceptionManager.getInstance().getFullMessage("bp_JETS_2");
      throw new BonitaRuntimeException(message);
    }
    if ((timer.getSignalName() == null) && (timer.getEventName() == null)) {
    	String message = ExceptionManager.getInstance().getFullMessage("bp_JETS_3");
    	throw new BonitaRuntimeException(message);
    }
    if (timer.getDueDate() == null) {
    	String message = ExceptionManager.getInstance().getFullMessage("bp_JETS_4");
    	throw new BonitaRuntimeException(message);
    }
    if (timer.getDueDate().getTime() < 0) {
    	String message = ExceptionManager.getInstance().getFullMessage("bp_JETS_5");
    	throw new BonitaRuntimeException(message);
    }
  }

  public void cancel(Timer timer) {
    log.debug("canceling " + timer);
    if (timer != null) {
      jobDbSession.delete(timer);
    } else {
    	String message = ExceptionManager.getInstance().getFullMessage("bp_JETS_6");
      throw new BonitaRuntimeException(message);
    }
  }
}
