package org.ow2.bonita.pvm.internal.wire.operation;

import org.ow2.bonita.env.Transaction;
import org.ow2.bonita.pvm.internal.tx.StandardResource;
import org.ow2.bonita.pvm.internal.tx.StandardTransaction;
import org.ow2.bonita.pvm.internal.wire.WireContext;
import org.ow2.bonita.pvm.internal.wire.WireException;
import org.ow2.bonita.util.ExceptionManager;
import org.ow2.bonita.util.Log;

/**
 * enlists this {@link StandardResource} with the current {@link Transaction}.
 * 
 * <p>
 * This {@link Operation} specifies that the object on which this operation is
 * applied should be added as a {@link StandardResource} to the specified
 * {@link Transaction}.
 * </p>
 * 
 * <p>
 * property transactionName refers to the objectName of the {@link Transaction}
 * and it may not be null.
 * </p>
 * 
 * @author Tom Baeyens
 * @author Guillaume Porcher (documentation)
 */
public class EnlistOperation implements Operation {

  private static final long serialVersionUID = 1L;
  private static Log log = Log.getLog(EnlistOperation.class.getName());

  String transactionName = null;

  /**
   * @throws WireException
   *           if this operation is applied on an object which is not a resource
   *           or if the specified transaction cannot be found.
   */
  public void apply(Object target, WireContext wireContext) {
    if (!(target instanceof StandardResource)) {
    	String message = ExceptionManager.getInstance().getFullMessage(
    			"bp_EO_1", StandardResource.class.getName(), target,
    			(target != null ? " (" + target.getClass().getName() + ")" : ""));
      throw new WireException(message);
    }

    Object object = null;
    if (transactionName != null) {
      object = wireContext.get(transactionName);
    } else {
      object = wireContext.get(Transaction.class);
    }

    if ((object == null) || (!(object instanceof StandardTransaction))) {
    	String message = ExceptionManager.getInstance().getFullMessage(
    			"bp_EO_2", StandardTransaction.class.getName(),
    			(transactionName != null ? "'" + transactionName + "'" : "by type"), target);
      throw new WireException(message);
    }

    StandardTransaction standardTransaction = (StandardTransaction) object;

    log.trace("enlisting resource " + target + " with transaction");
    standardTransaction.enlistResource((StandardResource) target);
  }

  /**
   * Gets the name of the transaction to which the object should be added.
   */
  public String getTransactionName() {
    return transactionName;
  }

  /**
   * Sets the name of the transaction to which the object should be added.
   * 
   * @param transactionName
   */
  public void setTransactionName(String transactionName) {
    this.transactionName = transactionName;
  }
}
