/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.runtime;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.bonita.definition.InternalProcess;
import org.ow2.bonita.definition.activity.AbstractActivity;
import org.ow2.bonita.facade.runtime.InstanceState;
import org.ow2.bonita.facade.uuid.ActivityInstanceUUID;
import org.ow2.bonita.pvm.Execution;
import org.ow2.bonita.pvm.internal.model.ExecutionImpl;
import org.ow2.bonita.pvm.internal.type.Variable;
import org.ow2.bonita.pvm.model.OpenExecution;
import org.ow2.bonita.services.Recorder;
import org.ow2.bonita.services.info.ActivityInstanceCurrentInfo;
import org.ow2.bonita.services.info.ProcessInstanceCurrentInfo;
import org.ow2.bonita.services.util.ServiceEnvTool;
import org.ow2.bonita.util.Misc;

/**
 * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes, Pierre Vigneras
 */

/**
 *  Variables in internalExecution are handled this way:
 *  - Process (global) variables are all in one VariableScope. This variable Scope always exists,
 *  even if it is empty.
 *  - Activity (local) variables are in a second variableScope. This variable scope only exists
 *  when there are local variables.
 *
 *  Methods getLocalVariables and getGlobalVariables are provided to have access to global and
 *  local variables of this execution.
 */
public class InternalExecution extends ExecutionImpl {

  private static final long serialVersionUID = 1L;

  private static final Logger LOG = Logger.getLogger(InternalExecution.class.getName());

  public static final String MAIN_INSTANCE_NAME = "mainInstance";
  protected InternalInstance instance;

  protected InternalInstance subflowInstance;

  protected String iterationId;
  protected String activityInstanceId = MAIN_INSTANCE_NAME;
  protected int waitingForActivityInstanceNb;


  protected ActivityInstanceUUID currentActivityInstanceUUID = null;

  // mandatory for persistence
  protected InternalExecution() {
    super();
  }

  @Override
  public InternalExecution getParent() {
    return (InternalExecution) this.parent;
  }

  @Override
  protected InternalExecution newChildExecution() {
    final InternalExecution child = new InternalExecution();
    child.instance = this.getInstance();
    child.iterationId = this.getIterationId();
    return child;
  }

  @Override
  public InternalExecution getProcessInstance() {
    return (InternalExecution) super.getProcessInstance();
  }

  @Override
  public InternalProcess getProcessDefinition() {
    return (InternalProcess) this.processDefinition;
  }

  @Override
  public void begin() {
    final InternalInstance instance = getInstance();
    instance.setInstanceState(InstanceState.STARTED);
    setIterationId(Misc.getUniqueId("it"));
    super.begin();
  }

  /* ITERATORS */

  public InternalInstance getInstance() {
    return this.instance;
  }

  public String getIterationId() {
    return this.iterationId;
  }

  public void setIterationId(final String iterationId) {
    this.iterationId = iterationId;
  }

  @SuppressWarnings("unchecked")
  public Map<String, Object> getScopeVariables() {
    if (this.variables == null) {
      return Collections.EMPTY_MAP;
    }
    final Map<String, Object> values = new HashMap<String, Object>();
    for (final Map.Entry<String, Variable> entry : this.variables.entrySet()) {
      final String name = entry.getKey();
      final Variable variable = entry.getValue();
      final Object value = variable.getValue();
      values.put(name, value);
    }
    return values;
  }

  public String getActivityInstanceId() {
    return this.activityInstanceId;
  }
  public void setActivityInstanceId(final String activityInstanceId) {
    this.activityInstanceId = activityInstanceId;
  }
  public int getWaitingForActivityInstanceNb() {
    return this.waitingForActivityInstanceNb;
  }
  public void setWaitingForActivityInstanceNb(final int waitingFor) {
    this.waitingForActivityInstanceNb = waitingFor;
  }
  public ActivityInstanceUUID getCurrentActivityInstanceUUID() {
    return this.currentActivityInstanceUUID;
  }
  public void setCurrentActivityInstanceUUID(final ActivityInstanceUUID currentActivityInstanceUUID) {
    this.currentActivityInstanceUUID = currentActivityInstanceUUID;
  }


  public InternalInstance getSubflowInstance() {
    return this.subflowInstance;
  }


  public void setSubflowInstance(final InternalInstance subflowInstance) {
    this.subflowInstance = subflowInstance;
  }

  public void abort() {
    if (this.getExecutions() != null) {
      for (final OpenExecution child : new ArrayList<OpenExecution>(this.getExecutions())) {
        ((InternalExecution) child).abort();
      }
    }
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine(this + " aborted.");
    }

    if (this.getCurrentActivityInstanceUUID() != null) {
      final AbstractActivity activity = (AbstractActivity) this.getNode().getBehaviour();
      boolean isSubflow = activity.getActivityDefiniton().isSubflow();
      if (isSubflow) {
        final InternalInstance childInstance = this.getSubflowInstance();
        childInstance.getRootExecution().abort();
        childInstance.setInstanceState(InstanceState.ABORTED);
        childInstance.setExecutionToSignal(null);
        final Recorder recorder = ServiceEnvTool.getRecorder();

        recorder.recordInstanceAborted(new ProcessInstanceCurrentInfo(
            childInstance.getProcessDefinitionUUID(),
            childInstance.getUUID(),
            childInstance.getParentInstance().getUUID(),
            childInstance.getRootExecution().getVariables(),
            ServiceEnvTool.getUserId())
        );
      }
      final ActivityInstanceCurrentInfo activityInstanceCurrentInfo = new ActivityInstanceCurrentInfo(
          this.getInstance().getProcessDefinitionUUID(),
          this.getInstance().getUUID(),
          this.getCurrentActivityInstanceUUID(),
          activity.getActivityDefiniton().isAutomatic(),
          activity.getActivityDefiniton().getName(),
          activity.getActivityDefiniton().getLabel(),
          activity.getActivityDefiniton().getDescription(),
          this.getIterationId(),
          this.getActivityInstanceId(),
          this.getScopeVariables());

      ServiceEnvTool.getRecorder().recordBodyAborted(activityInstanceCurrentInfo);
    }

    end(Execution.STATE_CANCELLED);
    final InternalExecution parent = this.getParent();
    if (parent != null) {
      parent.removeExecution(this);
    }
  }

  @Override
  public void cancel() {
    if (this.getExecutions() != null) {
      for (final OpenExecution child : new ArrayList<OpenExecution>(this.getExecutions())) {
        ((InternalExecution) child).cancel();
      }
    }
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine(this + " cancelled.");
    }

    if (this.getCurrentActivityInstanceUUID() != null) {
      final AbstractActivity activity = (AbstractActivity) this.getNode().getBehaviour();
      boolean isSubflow = activity.getActivityDefiniton().isSubflow();
      if (isSubflow) {
        final InternalInstance childInstance = this.getSubflowInstance();
        childInstance.cancel();
      }
      final ActivityInstanceCurrentInfo activityInstanceCurrentInfo = new ActivityInstanceCurrentInfo(
          this.getInstance().getProcessDefinitionUUID(),
          this.getInstance().getUUID(),
          this.getCurrentActivityInstanceUUID(),
          activity.getActivityDefiniton().isAutomatic(),
          activity.getActivityDefiniton().getName(),
          activity.getActivityDefiniton().getLabel(),
          activity.getActivityDefiniton().getDescription(),
          this.getIterationId(),
          this.getActivityInstanceId(),
          this.getScopeVariables());

      ServiceEnvTool.getRecorder().recordBodyCancelled(activityInstanceCurrentInfo);
    }

    end(Execution.STATE_CANCELLED);
    final InternalExecution parent = this.getParent();
    if (parent != null) {
      parent.removeExecution(this);
    }
  }
  
  @Override
  public void end() {
    super.end();
    
  }
}
