/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 * 
 * Modified by Matthieu Chaffotte - BonitaSoft S.A.
 **/
package org.ow2.bonita.runtime;

import java.util.HashSet;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.bonita.connector.core.Filter;
import org.ow2.bonita.connector.impl.mapper.PerformerAssignFilter;
import org.ow2.bonita.definition.PerformerAssign;
import org.ow2.bonita.definition.RoleMapper;
import org.ow2.bonita.definition.activity.AbstractActivity;
import org.ow2.bonita.facade.def.element.FilterDefinition;
import org.ow2.bonita.facade.def.element.HookDefinition;
import org.ow2.bonita.facade.def.element.RoleMapperDefinition;
import org.ow2.bonita.facade.def.majorElement.ActivityDefinition;
import org.ow2.bonita.facade.def.majorElement.ParticipantDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessFullDefinition;
import org.ow2.bonita.facade.exception.BonitaWrapperException;
import org.ow2.bonita.facade.exception.IllegalTaskStateException;
import org.ow2.bonita.facade.exception.RoleMapperInvocationException;
import org.ow2.bonita.facade.exception.TaskNotFoundException;
import org.ow2.bonita.facade.runtime.ActivityInstance;
import org.ow2.bonita.facade.runtime.ActivityState;
import org.ow2.bonita.facade.runtime.StateUpdate;
import org.ow2.bonita.facade.runtime.TaskInstance;
import org.ow2.bonita.facade.runtime.impl.ActivityFullInstanceImpl;
import org.ow2.bonita.facade.uuid.ActivityDefinitionUUID;
import org.ow2.bonita.facade.uuid.ActivityInstanceUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.services.Recorder;
import org.ow2.bonita.services.Repository;
import org.ow2.bonita.services.info.ActivityInstanceCurrentInfo;
import org.ow2.bonita.services.info.TaskCurrentInfo;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.bonita.util.EngineEnvTool;
import org.ow2.bonita.util.ExceptionManager;

/**
 * is one task instance that can be assigned to an user
 * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes, Pierre Vigneras
 */
public class TaskManager {

  private static final Logger LOG = Logger.getLogger(TaskManager.class.getName());

  private static ActivityFullInstanceImpl getTask(final ActivityInstanceUUID taskUUID) throws TaskNotFoundException {
    ActivityFullInstanceImpl task = (ActivityFullInstanceImpl) EngineEnvTool.getAllQueriers().getTaskInstance(taskUUID);
    if (task == null) {
      throw new TaskNotFoundException("bai_RAPII_19", taskUUID);
    }
    return task;
  }

  private static ActivityDefinition getActivityDefinition(ProcessDefinitionUUID processUUID, String activityName) {
    ActivityDefinition activity = (ActivityDefinition) EngineEnvTool.getAllQueriers().getActivity(new ActivityDefinitionUUID(processUUID, activityName));
    return activity;
  }

  private static InternalExecution getExecution(final ActivityInstanceUUID taskUUID) {
    Repository repository = EngineEnvTool.getRepository();
    return repository.getExecutionOnActivity(taskUUID.getProcessInstanceUUID(), taskUUID);
  }


  private static void assign(final ActivityInstance task, final java.util.Set<java.lang.String> candidates, final String userId) {
    ActivityInstanceUUID taskUUID = task.getUUID();
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("assigning task : " + taskUUID + " on activity " + task.getActivityName());
    }
    final Recorder recorder = EngineEnvTool.getRecorder();
    recorder.recordTaskAssigned(
        new TaskCurrentInfo(
            taskUUID,
            userId,
            candidates,
            EngineEnvTool.getUserId(),
            task.getState())
    );
  }

  public static void unAssign(ActivityInstanceUUID taskUUID) throws TaskNotFoundException {
    ActivityFullInstanceImpl task = getTask(taskUUID);
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Unassigning: " + task);
    }
    //record the assign
    assign(task, task.getTaskCandidates(), null);

    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Unassigned: " + task);
    }
  }

  public static void assign(ActivityInstanceUUID taskUUID, Set<String> candidates) throws TaskNotFoundException {
    ActivityFullInstanceImpl task = getTask(taskUUID);
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Assigning: " + task);
    }
    //record the assign
    assign(task, candidates, task.getTaskUser());
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Assigned: " + task);
    }
  }

  public static void assign(ActivityInstanceUUID taskUUID, String userId) throws TaskNotFoundException {
    ActivityFullInstanceImpl task = getTask(taskUUID);
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Assigning: " + task + " to " + userId);
    }
    //record the assign
    assign(task, task.getTaskCandidates(), userId);
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Assigned: " + task + " to " + userId);
    }
  }

  private static Set<String> getCandidates(Set<Performer> performers, TaskInstance task) {
    Set<String> candidates = null;
    if (performers != null) {
      candidates = new HashSet<String>();
      for (Performer performer : performers) {
        if (performer.isHuman()) {
          candidates.add(performer.getName());
        } else if (performer.getRoleMapperDefinition() != null) {
          Set<String> tmp = executeRoleMapper(task, performer);
          if (tmp != null) {
            candidates.addAll(tmp);
          }
        }
      }
    }
    return candidates;
  }

  public static void assign(ActivityInstanceUUID taskUUID) throws TaskNotFoundException {
    ActivityFullInstanceImpl task = getTask(taskUUID);
    ProcessDefinitionUUID processUUID = task.getProcessDefinitionUUID();
    ActivityDefinition activityDefinition = getActivityDefinition(processUUID, task.getActivityName());

    //execute role resolvers and filters
    Set<Performer> performers = getPerformers(activityDefinition);

    Set<String> candidates = getCandidates(performers, task);
    String userId = null;

    Performer performer = performers.iterator().next();
    if (performer.getFilterDefinition() != null) {
      candidates = executeFilter(task, candidates, performer);
      if (candidates.size() == 1) {
        userId = candidates.iterator().next();
      }
    }

    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Assigning: " + task);
    }
    //record the assign
    assign(task, candidates, userId);
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Assigned: " + task);
    }
  }


  public static void finish(final ActivityInstanceUUID taskUUID, final boolean assignTask) throws TaskNotFoundException, IllegalTaskStateException {
    ActivityFullInstanceImpl task = getTask(taskUUID);

    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Checking compatible state of " + task);
    }
    if (!task.getState().equals(ActivityState.EXECUTING)) {
      final Set<ActivityState> expectedStates = new HashSet<ActivityState>();
      expectedStates.add(ActivityState.EXECUTING);
      String message = ExceptionManager.getInstance().getFullMessage("bai_RAPII_11");
      throw new IllegalTaskStateException("bai_RAPII_11", message,
          taskUUID, expectedStates, task.getState());
    }
    String activityName = task.getActivityName();
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Finishing task : " + taskUUID + " on activity " + activityName);
    }
    final Recorder recorder = EngineEnvTool.getRecorder();
    if (assignTask) {
      assign(task, task.getTaskCandidates(), task.getTaskUser());
    }
    recorder.recordTaskFinished(
        new TaskCurrentInfo(
            task.getUUID(),
            task.getTaskUser(),
            task.getTaskCandidates(),
            EngineEnvTool.getUserId(),
            ActivityState.FINISHED)
    );

    InternalExecution internalExecution = getExecution(taskUUID);
    AbstractActivity abstractActivity = (AbstractActivity) internalExecution.getNode().getBehaviour();

    EngineEnvTool.getHookExecutor().executeConnectors(abstractActivity.getActivityDefiniton(), internalExecution, HookDefinition.Event.taskOnFinish);
    abstractActivity.signal(internalExecution, AbstractActivity.BODY_FINISHED, null);    

    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Terminated: " + task);
    }
  }

  public static void suspend(final ActivityInstanceUUID taskUUID, final boolean assignTask) throws TaskNotFoundException, IllegalTaskStateException {
    ActivityFullInstanceImpl task = getTask(taskUUID);

    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Checking compatible state of " + task);
    }
    if (!(task.getState().equals(ActivityState.READY)
        || task.getState().equals(ActivityState.EXECUTING))) {
      final Set<ActivityState> expectedStates = new HashSet<ActivityState>();
      expectedStates.add(ActivityState.READY);
      expectedStates.add(ActivityState.EXECUTING);

      String message = ExceptionManager.getInstance().getFullMessage("bai_RAPII_13");
      throw new IllegalTaskStateException("bai_RAPII_13", message,
          taskUUID, expectedStates, task.getState());
    }

    final String currentUserId = EngineEnvTool.getUserId();

    ProcessDefinitionUUID processUUID = task.getProcessDefinitionUUID();
    String activityName = task.getActivityName();

    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Suspending task : " + taskUUID + " on activity " + activityName);
    }
    final Recorder recorder = EngineEnvTool.getRecorder();
    if (assignTask) {
      assign(task, task.getTaskCandidates(), currentUserId);
    }
    recorder.recordTaskSuspended(
        new TaskCurrentInfo(
            taskUUID,
            task.getTaskUser(),
            task.getTaskCandidates(),
            EngineEnvTool.getUserId(),
            ActivityState.SUSPENDED)
    );

    InternalExecution internalExecution = getExecution(taskUUID);

    EngineEnvTool.getHookExecutor().executeConnectors(getActivityDefinition(processUUID, activityName), internalExecution, HookDefinition.Event.taskOnSuspend);

    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Suspended: " + task);
    }
  }

  public static void resume(final ActivityInstanceUUID taskUUID, final boolean taskAssign) throws IllegalTaskStateException, TaskNotFoundException {
    ActivityFullInstanceImpl task = getTask(taskUUID);

    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Checking compatible state of " + task);
    }
    if (!task.getState().equals(ActivityState.SUSPENDED)) {
      final Set<ActivityState> expectedStates = new HashSet<ActivityState>();
      expectedStates.add(ActivityState.SUSPENDED);
      String message = ExceptionManager.getInstance().getFullMessage("bai_RAPII_15");
      throw new IllegalTaskStateException("bai_RAPII_15", message,
          taskUUID, expectedStates, task.getState());
    }
    final String currentUserId = EngineEnvTool.getUserId();

    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Resuming: " + task);
    }

    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Resuming task : " + taskUUID + " on activity " + taskUUID.getActivityName());
    }
    final Recorder recorder = EngineEnvTool.getRecorder();
    if (taskAssign) {
      assign(task, task.getTaskCandidates(), currentUserId);
    }
    ActivityState stateBeforeSuspend = null;
    for (StateUpdate su : task.getStateUpdates()) {
      if (su.getActivityState().equals(ActivityState.SUSPENDED)) {
        stateBeforeSuspend = su.getInitialState();
      }
    }
    recorder.recordTaskResumed(
        new TaskCurrentInfo(
            taskUUID,
            task.getTaskUser(),
            task.getTaskCandidates(),
            EngineEnvTool.getUserId(),
            stateBeforeSuspend)
    );

    InternalExecution internalExecution = getExecution(taskUUID);
    AbstractActivity abstractActivity = (AbstractActivity) internalExecution.getNode().getBehaviour();

    EngineEnvTool.getHookExecutor().executeConnectors(abstractActivity.getActivityDefiniton(), internalExecution, HookDefinition.Event.taskOnResume);

    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Resumed: " + task);
    }
  }

  public static void start(final ActivityInstanceUUID taskUUID, final boolean assignTask) throws IllegalTaskStateException, TaskNotFoundException {
    ActivityFullInstanceImpl task = getTask(taskUUID);

    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Checking compatible state of " + task);
    }
    if (!task.getState().equals(ActivityState.READY)) {
      final Set<ActivityState> expectedStates = new HashSet<ActivityState>();
      expectedStates.add(ActivityState.READY);
      String message = ExceptionManager.getInstance().getFullMessage("bai_RAPII_9");
      throw new IllegalTaskStateException("bai_RAPII_9", message, taskUUID, expectedStates, task.getState());
    }
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Starting: " + task);
    }

    final Recorder recorder = EngineEnvTool.getRecorder();
    if (assignTask) {
      assign(task, task.getTaskCandidates(), EngineEnvTool.getUserId());
    }

    InternalExecution internalExecution = getExecution(taskUUID);
    AbstractActivity abstractActivity = (AbstractActivity) internalExecution.getNode().getBehaviour();

    final ActivityInstanceCurrentInfo activityInstanceCurrentInfo = new ActivityInstanceCurrentInfo(
        taskUUID.getProcessInstanceUUID().getProcessDefinitionUUID(),
        taskUUID.getProcessInstanceUUID(),
        taskUUID,
        false,
        taskUUID.getActivityName(),
        abstractActivity.getActivityDefiniton().getLabel(),
        abstractActivity.getActivityDefiniton().getDescription(),
        internalExecution.getIterationId(),
        internalExecution.getActivityInstanceId(),
        internalExecution.getScopeVariables());

    recorder.recordBodyStarted(activityInstanceCurrentInfo);
    recorder.recordTaskStarted(
        new TaskCurrentInfo(
            taskUUID,
            task.getTaskUser(),
            task.getTaskCandidates(),
            EngineEnvTool.getUserId(),
            task.getState())
    );

    EngineEnvTool.getHookExecutor().executeConnectors(abstractActivity.getActivityDefiniton(), internalExecution, HookDefinition.Event.taskOnStart);

    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Started: " + task);
    }
  }
  /*
  public static void abort(InternalExecution internalExecution) throws TaskNotFoundException {
    ActivityInstanceUUID taskUUID = new ActivityInstanceUUID(internalExecution.getCurrentActivityInstanceUUID().toString());

    ActivityFullInstanceImpl task = getTask(taskUUID);

    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Aborting task : " + taskUUID + " on activity " + taskUUID.getActivityName());
    }
    task.setActivityState(ActivityState.ABORTED, BonitaConstants.SYSTEM_USER);
  }

  public static void cancel(InternalExecution internalExecution) throws TaskNotFoundException {
    ActivityInstanceUUID taskUUID = new ActivityInstanceUUID(internalExecution.getCurrentActivityInstanceUUID().toString());
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Cancelling task : " + taskUUID + " on activity " + taskUUID.getActivityName());
    }

    ActivityFullInstanceImpl task = getTask(taskUUID);
    task.setActivityState(ActivityState.CANCELLED, BonitaConstants.SYSTEM_USER);
  }
   */

  private static Set<String> executeFilter(final TaskInstance task, final Set<String> candidates, final Performer performer) {
    final ProcessDefinitionUUID processUUID = task.getProcessDefinitionUUID();
    final FilterDefinition filterDefinition = performer.getFilterDefinition();
    if (filterDefinition != null) {
      PerformerAssign performerAssign = getPerformerAssign(processUUID, filterDefinition);
      Filter filter = null;
      if (performerAssign == null) {
        filter = ClassDataLoader.getInstance(Filter.class, processUUID, filterDefinition);
      } else {
        filter = new PerformerAssignFilter();
      }
      try {
        return EngineEnvTool.getHookExecutor().executeFilter(filter, performerAssign, task, candidates, filterDefinition.getParameters());

      } catch (Exception e) {
        throw new BonitaWrapperException(e);
      }
    }
    return candidates;
  }

  private static PerformerAssign getPerformerAssign(ProcessDefinitionUUID processUUID, FilterDefinition filterDefinition) {
    try {
      return ClassDataLoader.getInstance(PerformerAssign.class, processUUID, filterDefinition);
    } catch (Exception e) {
      return null;
    }
  }

  private static Set<String> executeRoleMapper(final TaskInstance task, final Performer performer) {
    ProcessDefinitionUUID processUUID = task.getProcessDefinitionUUID();
    final RoleMapperDefinition rolemapperDef = performer.getRoleMapperDefinition();
    if (rolemapperDef != null) {
      final RoleMapper roleMapper = ClassDataLoader.getInstance(RoleMapper.class, processUUID, rolemapperDef);
      try {
        return  EngineEnvTool.getHookExecutor().executeRoleMapper( roleMapper, task.getProcessInstanceUUID(), performer.getName(), rolemapperDef.getParameters());
      } catch (final Exception e) {
        throw new BonitaWrapperException(
            new RoleMapperInvocationException("be_TRT_2", rolemapperDef.toString(), e)
        );
      }
    }
    return null;
  }

  public static void ready(InternalExecution internalExecution) throws TaskNotFoundException {
    ActivityInstanceUUID taskUUID = new ActivityInstanceUUID(internalExecution.getCurrentActivityInstanceUUID().toString());
    AbstractActivity abstractActivity = (AbstractActivity) internalExecution.getNode().getBehaviour();
    ActivityDefinition activityDef = abstractActivity.getActivityDefiniton();

    ActivityFullInstanceImpl task = getTask(taskUUID);
    String activityName = taskUUID.getActivityName();
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Ready task : " + taskUUID + " on activity " + activityName);
    }

    // get performer (aka participant/swimlane)
    final Set<Performer> performers = getPerformers(activityDef);
    String userId = null;
    Set<String> candidates = null;
    if (performers != null) {
      candidates = getCandidates(performers, task);
      Performer performer = performers.iterator().next();
      if (performer.getFilterDefinition() != null) {
        candidates = executeFilter(task, candidates, performer);
        if (candidates.size() == 1) {
          userId = candidates.iterator().next();
        }
      }
    }

    EngineEnvTool.getRecorder().recordTaskReady(
        new TaskCurrentInfo(
            task.getUUID(),
            userId,
            candidates,
            EngineEnvTool.getUserId(),
            ActivityState.READY)
    );



    EngineEnvTool.getHookExecutor().executeConnectors(activityDef, internalExecution, HookDefinition.Event.taskOnReady);
  }

  protected static Performer getPerformer(final ActivityDefinition activity, final ProcessFullDefinition processDef, String performer) {
    ParticipantDefinition participant = null;
    if (processDef.getParticipants() != null) {
      for (final ParticipantDefinition p : processDef.getParticipants()) {
        if (p.getName().equals(performer)) {
          participant = p;
          break;
        }
      }
    }
    if (participant == null) {
      throw new BonitaRuntimeException("Wrong performer: " + performer
          + ". No participant is defined within the process with processDefinitionUUID: " + performer);
    }
    return new Performer(performer, participant.getRoleMapper(), activity.getFilter());
  }

  private static Set<Performer> getPerformers(ActivityDefinition activityDefinition) {
    Set<String> performers = activityDefinition.getPerformers();
    Set<Performer> result = null;
    if (performers != null && !performers.isEmpty()) {
      ProcessFullDefinition processDef = EngineEnvTool.getJournalQueriers().getProcess(activityDefinition.getProcessDefinitionUUID());
      result = new HashSet<Performer>();
      for (String performer : performers) {
        result.add(getPerformer(activityDefinition, processDef, performer));
      }
    }
    return result;
  }
  private static class Performer {
    protected String name;
    protected RoleMapperDefinition roleMapper;
    protected FilterDefinition filter;

    public Performer(final String name, final RoleMapperDefinition roleMapper, final FilterDefinition filter) {
      super();
      this.name = name;
      this.roleMapper = roleMapper;
      this.filter = filter;
    }

    public RoleMapperDefinition getRoleMapperDefinition() {
      return roleMapper;
    }
    public FilterDefinition getFilterDefinition() {
      return filter;
    }
    public String getName() {
      return this.name;
    }
    public boolean isHuman() {
      return roleMapper == null && filter == null;
    }
  }
}
